# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Anna Elisa Lappe
#
# Credit:
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# -------------------------------------------------------------------------------------

from argparse import Namespace
from pathlib import Path
from unittest.mock import ANY, MagicMock, call, patch

import pytest
from lightning.pytorch.callbacks.model_checkpoint import ModelCheckpoint

from itwinai.loggers import Logger
from itwinai.torch.loggers import ItwinaiLogger as PyTorchLightningLogger


def test_mock_experiment_log(lightning_mock_loggers):
    itwinai_logger_mock, lightning_logger = lightning_mock_loggers
    lightning_logger.experiment.log(
        item={"test": 10}, identifier="test_dict", kind="dict", step=1, batch_idx=1
    )
    # Check that the experiment properly initializes
    itwinai_logger_mock.create_logger_context.assert_called_once_with(rank=0)
    assert isinstance(lightning_logger.experiment, Logger)

    # Check that the log function is called properly
    itwinai_logger_mock.log.assert_called_once_with(
        item={"test": 10}, identifier="test_dict", kind="dict", step=1, batch_idx=1
    )


def test_mock_log_metrics(lightning_mock_loggers):
    """Test log_metrics method to ensure it logs metrics with the correct parameters."""
    itwinai_logger_mock, lightning_logger = lightning_mock_loggers
    metrics = {"accuracy": 0.95, "loss": 0.1}
    step = 1
    lightning_logger.log_metrics(metrics=metrics, step=step)
    # Check that the log function is called correctly
    for key, value in metrics.items():
        itwinai_logger_mock.log.assert_any_call(
            item=value, identifier=key, kind="metric", step=step
        )


def test_moc_log_hyperparams(lightning_mock_loggers):
    """Test log_hyperparams method to ensure it logs hyperparameters correctly."""
    itwinai_logger_mock, lightning_logger = lightning_mock_loggers

    dict_params = {"learning_rate": 0.001, "batch_size": 32}
    lightning_logger.log_hyperparams(params=dict_params)
    itwinai_logger_mock.save_hyperparameters.assert_called_once_with(dict_params)

    itwinai_logger_mock.reset_mock()

    namespace_params = Namespace(**dict_params)
    lightning_logger.log_hyperparams(params=namespace_params)
    itwinai_logger_mock.save_hyperparameters.assert_called_once_with(dict_params)


def test_mock_after_save_checkpoint(lightning_mock_loggers):
    """Test after_save_checkpoint method in PyTorchLightningLogger."""
    _, lightning_logger = lightning_mock_loggers

    checkpoint_callback = MagicMock(spec=ModelCheckpoint)
    checkpoint_callback.save_top_k = -1
    checkpoint_callback.best_model_path = Path("/path/to/best_model.ckpt")
    assert lightning_logger._checkpoint_callback is None

    # Scenario 1:
    lightning_logger._log_model = "all"
    lightning_logger.after_save_checkpoint(checkpoint_callback)
    assert lightning_logger._checkpoint_callback is None

    # Scenario 2: Should checkpoint
    lightning_logger._log_model = True
    checkpoint_callback.save_top_k = -1
    lightning_logger.after_save_checkpoint(checkpoint_callback)
    assert lightning_logger._checkpoint_callback is None

    # Scenario 3:
    lightning_logger._log_model = False
    lightning_logger.after_save_checkpoint(checkpoint_callback)
    assert lightning_logger._checkpoint_callback is None

    # Scenario 4:
    lightning_logger._log_model = True
    checkpoint_callback.save_top_k = 5
    lightning_logger.after_save_checkpoint(checkpoint_callback)
    assert lightning_logger._checkpoint_callback == checkpoint_callback


@pytest.mark.parametrize(
    "itwinai_logger",
    [
        "console_logger",
        "mlflow_logger",
        "wandb_logger",
        "tensorboard_logger_torch",
        "prov4ml_logger",
        "loggers_collection",
    ],
)
def test_experiment_and_finalize(itwinai_logger, request):
    itwinai_logger_instance = request.getfixturevalue(itwinai_logger)
    lightning_logger = PyTorchLightningLogger(itwinai_logger=itwinai_logger_instance)

    with patch.object(itwinai_logger_instance, "create_logger_context"):
        assert not lightning_logger._initialized
        assert lightning_logger.name is None
        assert lightning_logger.version is None

        experiment = lightning_logger.experiment
        itwinai_logger_instance.create_logger_context.assert_called_once_with(rank=0)
        assert isinstance(experiment, Logger)


@pytest.mark.parametrize(
    "itwinai_logger",
    [
        "console_logger",
        "mlflow_logger",
        "wandb_logger",
        "tensorboard_logger_torch",
        "prov4ml_logger",
        "loggers_collection",
    ],
)
def test_log_metrics_and_hyperparams(itwinai_logger, request):
    itwinai_logger_instance = request.getfixturevalue(itwinai_logger)
    lightning_logger = PyTorchLightningLogger(itwinai_logger=itwinai_logger_instance)

    with (
        patch.object(itwinai_logger_instance, "create_logger_context"),
        patch.object(itwinai_logger_instance, "log"),
        patch.object(itwinai_logger_instance, "save_hyperparameters"),
    ):
        lightning_logger.log_metrics({"loss": 0.2, "accuracy": 0.99})
        expected_calls = [
            call(item=0.2, identifier="loss", kind="metric", step=None),
            call(item=0.99, identifier="accuracy", kind="metric", step=None),
        ]
        itwinai_logger_instance.log.assert_has_calls(expected_calls)

        dict_params = {"learning_rate": 0.001, "batch_size": 32}
        lightning_logger.log_hyperparams(params=dict_params)
        itwinai_logger_instance.save_hyperparameters.assert_called_once_with(dict_params)


@pytest.mark.parametrize(
    "itwinai_logger",
    [
        "console_logger",
        "mlflow_logger",
        "wandb_logger",
        "tensorboard_logger_torch",
        "prov4ml_logger",
        "loggers_collection",
    ],
)
def test_save_after_checkpoint(itwinai_logger, request):
    itwinai_logger_instance = request.getfixturevalue(itwinai_logger)
    lightning_logger = PyTorchLightningLogger(
        itwinai_logger=itwinai_logger_instance, log_model="all"
    )
    checkpoint_callback = MagicMock(spec=ModelCheckpoint)
    checkpoint_callback.best_model_path = Path("/path/to/checkpoint_1.ckpt")

    with (
        patch.object(itwinai_logger_instance, "create_logger_context"),
        patch.object(itwinai_logger_instance, "log"),
        patch(
            "itwinai.torch.loggers._scan_checkpoints",
            return_value=[
                (1, "path/to/checkpoint_1.ckpt", 100, "_"),
                (2, "path/to/checkpoint_2.ckpt", 101, "_"),
            ],
        ),
    ):
        lightning_logger.after_save_checkpoint(checkpoint_callback)
        assert lightning_logger._checkpoint_callback is None

        expected_calls = [
            call(
                item="path/to/checkpoint_1.ckpt",
                identifier="checkpoint_1",
                kind="artifact",
            ),
            call(item=ANY, identifier="checkpoint_1", kind="artifact"),
            call(
                item="path/to/checkpoint_2.ckpt",
                identifier="checkpoint_2",
                kind="artifact",
            ),
            call(item=ANY, identifier="checkpoint_2", kind="artifact"),
        ]
        itwinai_logger_instance.log.assert_has_calls(expected_calls)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# -------------------------------------------------------------------------------------

import shutil
from unittest.mock import MagicMock

import pytest

from itwinai.loggers import (
    ConsoleLogger,
    Logger,
    LoggersCollection,
    MLFlowLogger,
    Prov4MLLogger,
    TensorBoardLogger,
    WandBLogger,
)
from itwinai.torch.loggers import ItwinaiLogger as PyTorchLightningLogger


@pytest.fixture(scope="module")
def console_logger():
    yield ConsoleLogger(savedir="/tmp/console/test_mllogs", log_freq=1)
    shutil.rmtree("/tmp/console/test_mllogs", ignore_errors=True)


@pytest.fixture(scope="module")
def mlflow_logger():
    yield MLFlowLogger(
        savedir="/tmp/mlflow/test_mllogs",
        experiment_name="test_experiment",
        tracking_uri="file:///tmp/mlruns",
    )
    shutil.rmtree("/tmp/mlflow/test_mllogs", ignore_errors=True)
    shutil.rmtree("/tmp/mlruns", ignore_errors=True)


@pytest.fixture(scope="module")
def wandb_logger():
    yield WandBLogger(savedir="/tmp/wandb/test_mllogs", project_name="test_project")
    shutil.rmtree("/tmp/wandb/test_mllogs", ignore_errors=True)


@pytest.fixture(scope="module")
def tensorboard_logger_tf():
    yield TensorBoardLogger(savedir="/tmp/tf_tb/test_mllogs", framework="tensorflow")
    shutil.rmtree("/tmp/tf_tb/test_mllogs", ignore_errors=True)


@pytest.fixture(scope="module")
def tensorboard_logger_torch():
    yield TensorBoardLogger(savedir="/tmp/torch_tb/test_mllogs", framework="pytorch")
    shutil.rmtree("/tmp/torch_tb/test_mllogs", ignore_errors=True)


@pytest.fixture(scope="module")
def prov4ml_logger():
    return Prov4MLLogger()


@pytest.fixture(scope="module")
def loggers_collection(
    console_logger,
    mlflow_logger,
    wandb_logger,
    tensorboard_logger_torch,
    prov4ml_logger,
):
    return LoggersCollection(
        [
            console_logger,
            mlflow_logger,
            wandb_logger,
            tensorboard_logger_torch,
            prov4ml_logger,
        ]
    )


@pytest.fixture
def lightning_mock_loggers():
    """Setup a generic PyTorchLightningLogger with mock loggers for testing."""
    itwinai_logger_mock = MagicMock(spec=Logger)
    lightning_logger = PyTorchLightningLogger(itwinai_logger=itwinai_logger_mock)

    return itwinai_logger_mock, lightning_logger



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Henry Mutegeki <henry.mutegeki@cern.ch> - CERN
# - Iacopo Ferrario <iacopofederico.ferrario@eurac.edu> - EURAC
# --------------------------------------------------------------------------------------

from typing import Any, Tuple

import xarray as xr
from hython.datasets import LSTMDataset, get_dataset
from hython.io import read_from_zarr
from hython.normalizer import Normalizer
from hython.sampler import AbstractDownSampler

from itwinai.components import DataProcessor, DataSplitter, monitor_exec


class RNNDatasetGetterAndSplitter(DataSplitter):
    def __init__(
        self,
        surrogate_input: str,
        dynamic_names: list[str],
        static_names: list[str],
        target_names: list[str],
        mask_names: list[str],
        train_temporal_range: list[str] = ["", ""],
        test_temporal_range: list[str] = ["", ""],
        name: str | None = None,
    ) -> None:
        self.save_parameters(**self.locals2params(locals()))
        self.surrogate_input = surrogate_input
        self.dynamic_names = dynamic_names
        self.static_names = static_names
        self.target_names = target_names
        self.mask_names = mask_names
        self.train_temporal_range = slice(train_temporal_range[0], train_temporal_range[1])
        self.test_temporal_range = slice(test_temporal_range[0], test_temporal_range[1])

    @monitor_exec
    def execute(self) -> Tuple:
        # Dataset preparation
        Xd = (
            read_from_zarr(url=self.surrogate_input, group="xd", multi_index="gridcell")
            .sel(time=self.train_temporal_range)
            .xd.sel(feat=self.dynamic_names)
        )
        Xs = read_from_zarr(
            url=self.surrogate_input, group="xs", multi_index="gridcell"
        ).xs.sel(feat=self.static_names)
        Y = (
            read_from_zarr(url=self.surrogate_input, group="y", multi_index="gridcell")
            .sel(time=self.train_temporal_range)
            .y.sel(feat=self.target_names)
        )

        Y_test = (
            read_from_zarr(url=self.surrogate_input, group="y", multi_index="gridcell")
            .sel(time=self.test_temporal_range)
            .y.sel(feat=self.target_names)
        )
        Xd_test = (
            read_from_zarr(url=self.surrogate_input, group="xd", multi_index="gridcell")
            .sel(time=self.test_temporal_range)
            .xd.sel(feat=self.dynamic_names)
        )

        masks = (
            read_from_zarr(url=self.surrogate_input, group="mask")
            .mask.sel(mask_layer=self.mask_names)
            .any(dim="mask_layer")
        )
        # pass to rnnprocessor
        return (Xd, Xs, Y), (Xd_test, Xs, Y_test), None, masks


class RNNProcessor(DataProcessor):
    def __init__(
        self,
        dataset: str,
        downsampling_train: AbstractDownSampler = None,
        downsampling_test: AbstractDownSampler = None,
        normalizer_dynamic: Normalizer = None,
        normalizer_static: Normalizer = None,
        normalizer_target: Normalizer = None,
        name: str | None = None,
    ) -> None:
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))
        self.dataset = dataset
        self.downsampling_train = downsampling_train
        self.downsampling_test = downsampling_test
        # For the moment these are fixed
        self.normalizer_dynamic = Normalizer(
            method="standardize", type="spacetime", axis_order="NTC"
        )
        self.normalizer_static = Normalizer(
            method="standardize", type="space", axis_order="NTC"
        )
        self.normalizer_target = Normalizer(
            method="standardize", type="spacetime", axis_order="NTC"
        )

    @monitor_exec
    def execute(
        self,
        train_dataset: Tuple,
        validation_dataset: Tuple,
        test_dataset: Any = None,
        masks: xr.Dataset = None,
    ) -> Tuple[LSTMDataset, LSTMDataset, None]:
        Xd, Xs, Y = train_dataset
        Xd_test, Xs_test, Y_test = validation_dataset

        SHAPE = Xd.attrs["shape"]

        train_dataset = get_dataset(self.dataset)(
            Xd,
            Y,
            Xs,
            original_domain_shape=SHAPE,
            mask=masks,
            downsampler=self.downsampling_train,
            normalizer_dynamic=self.normalizer_dynamic,
            normalizer_static=self.normalizer_static,
            normalizer_target=self.normalizer_target,
        )
        validation_dataset = get_dataset(self.dataset)(
            Xd_test,
            Y_test,
            Xs_test,
            original_domain_shape=SHAPE,
            mask=masks,
            downsampler=self.downsampling_test,
            normalizer_dynamic=self.normalizer_dynamic,
            normalizer_static=self.normalizer_static,
            normalizer_target=self.normalizer_target,
        )

        # Pass to trainer
        return train_dataset, validation_dataset, None


class ConvRNNDatasetGetterAndSplitter(DataSplitter):
    def __init__(
        self,
        surrogate_input: str,
        dynamic_names: list[str],
        static_names: list[str],
        target_names: list[str],
        mask_names: list[str],
        train_temporal_range: list[str] = ["", ""],
        test_temporal_range: list[str] = ["", ""],
        name: str | None = None,
    ) -> None:
        self.save_parameters(**self.locals2params(locals()))
        self.surrogate_input = surrogate_input
        self.dynamic_names = dynamic_names
        self.static_names = static_names
        self.target_names = target_names
        self.mask_names = mask_names
        self.train_temporal_range = slice(train_temporal_range[0], train_temporal_range[1])
        self.test_temporal_range = slice(test_temporal_range[0], test_temporal_range[1])

    @monitor_exec
    def execute(self) -> Tuple:
        # Dataset preparation
        Xd = read_from_zarr(url=self.surrogate_input, group="xd").sel(
            time=self.train_temporal_range
        )[self.dynamic_names]
        Xs = read_from_zarr(url=self.surrogate_input, group="xs")[self.static_names]
        Y = read_from_zarr(url=self.surrogate_input, group="y").sel(
            time=self.train_temporal_range
        )[self.target_names]

        Y_test = read_from_zarr(url=self.surrogate_input, group="y").sel(
            time=self.test_temporal_range
        )[self.target_names]

        Xd_test = read_from_zarr(url=self.surrogate_input, group="xd").sel(
            time=self.test_temporal_range
        )[self.dynamic_names]
        masks = (
            read_from_zarr(url=self.surrogate_input, group="mask")
            .mask.sel(mask_layer=self.mask_names)
            .any(dim="mask_layer")
        )
        # pass to rnnprocessor
        return (Xd, Xs, Y), (Xd_test, Xs, Y_test), None, masks


class ConvRNNProcessor(DataProcessor):
    def __init__(
        self,
        dataset: str,
        downsampling_train: AbstractDownSampler = None,
        downsampling_test: AbstractDownSampler = None,
        normalizer_dynamic: Normalizer = None,
        normalizer_static: Normalizer = None,
        normalizer_target: Normalizer = None,
        name: str | None = None,
    ) -> None:
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))
        self.dataset = dataset
        self.downsampling_train = downsampling_train
        self.downsampling_test = downsampling_test
        # For the moment these are fixed
        self.normalizer_dynamic = Normalizer(
            method="standardize", type="spacetime", axis_order="xarray_dataset"
        )
        self.normalizer_static = Normalizer(
            method="standardize", type="space", axis_order="xarray_dataset"
        )
        self.normalizer_target = Normalizer(
            method="standardize", type="spacetime", axis_order="xarray_dataset"
        )

        # TODO: HARDCODED!
        XSIZE, YSIZE, TSIZE = 20, 20, 60
        XOVER, YOVER, TOVER = 5, 5, 50
        self.missing_policy = 0.05
        self.batch_size = {"xsize": XSIZE, "ysize": YSIZE, "tsize": TSIZE}
        self.overlap = {"xover": XOVER, "yover": YOVER, "tover": TOVER}
        self.persist = True
        self.static_to_dynamic = True

    @monitor_exec
    def execute(
        self,
        train_dataset: Tuple,
        validation_dataset: Tuple,
        test_dataset: Any = None,
        masks: xr.Dataset = None,
    ) -> Tuple[LSTMDataset, LSTMDataset, None]:
        Xd, Xs, Y = train_dataset
        Xd_test, Xs_test, Y_test = validation_dataset

        SHAPE = Xd.precip.shape

        train_dataset = get_dataset(self.dataset)(
            Xd,
            Y,
            Xs,
            mask=masks,
            downsampler=self.downsampling_train,
            normalizer_dynamic=self.normalizer_dynamic,
            normalizer_static=self.normalizer_static,
            normalizer_target=self.normalizer_target,
            shape=SHAPE,
            batch_size=self.batch_size,
            overlap=self.overlap,
            missing_policy=self.missing_policy,
            persist=self.persist,
            static_to_dynamic=self.static_to_dynamic,
        )
        validation_dataset = get_dataset(self.dataset)(
            Xd_test,
            Y_test,
            Xs_test,
            mask=masks,
            downsampler=self.downsampling_test,
            normalizer_dynamic=self.normalizer_dynamic,
            normalizer_static=self.normalizer_static,
            normalizer_target=self.normalizer_target,
            shape=SHAPE,
            batch_size=self.batch_size,
            overlap=self.overlap,
            missing_policy=self.missing_policy,
            persist=self.persist,
            static_to_dynamic=self.static_to_dynamic,
        )

        # Pass to trainer
        return train_dataset, validation_dataset, None



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import pytest
from pydantic import ValidationError

from itwinai.torch.config import TrainingConfiguration


def test_values_parsing():
    """Check dynamic override and creation of new entries."""
    cfg = TrainingConfiguration(batch_size="11", param_abc="11", param_xyz=1.1)
    assert cfg.batch_size == 11
    assert cfg.param_abc == "11"
    assert cfg.param_xyz == 1.1
    assert isinstance(cfg.pin_gpu_memory, bool)

    # Check dict-like getitem
    assert cfg["batch_size"] == 11


def test_illegal_override():
    """Test that illegal type override fails."""
    with pytest.raises(ValidationError) as exc_info:
        TrainingConfiguration(batch_size="hello")
    assert "batch_size" in str(exc_info.value)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
#
# --------------------------------------------------------------------------------------
# Command-line interface for the itwinai Python library.
# Example:
#
# >>> itwinai --help
#
# --------------------------------------------------------------------------------------
#
# NOTE: import libraries in the command's function, not here, as having them here will
# slow down the CLI commands significantly.

from pathlib import Path
from typing import List, Optional

import typer
from typing_extensions import Annotated

app = typer.Typer(pretty_exceptions_enable=False)


@app.command()
def generate_gpu_data_plots(
    log_dir: str = "scalability-metrics/gpu-energy-data",
    pattern: str = r".*\.csv$",
    plot_dir: str = "plots/",
    do_backup: bool = False,
    backup_dir: str = "backup-scalability-metrics/",
    experiment_name: Optional[str] = None,
    run_name: Optional[str] = None,
) -> None:
    """Generate GPU energy and utilization plots showing the expenditure for each
    combination of strategy and number of GPUs in Watt hours and total computing
    percentage. Backs up the data used to create the plot if ``backup_dir`` is not None

    Args:
        log_dir: The directory where the csv logs are stored. Defaults to
            ``utilization_logs``.
        pattern: A regex pattern to recognize the file names in the 'log_dir' folder.
            Defaults to ``dataframe_(?:\\w+)_(?:\\d+)\\.csv$``. Set it to 'None' to
            make it None. In this case, it will match all files in the given folder.
        plot_dir: The directory where the resulting plots should be saved. Defaults to
            ``plots/``.
        do_backup: Whether to backup the data used for making the plot or not.
        backup_dir: The path to where the data used to produce the plot should be
            saved.
        experiment_name: The name of the experiment to be used when creating a backup
            of the data used for the plot.
        run_name: The name of the run to be used when creating a backup of the data
            used for the plot.

    """

    from itwinai.scalability import (
        backup_scalability_metrics,
        convert_matching_files_to_dataframe,
    )
    from itwinai.torch.monitoring.plotting import (
        calculate_average_gpu_utilization,
        calculate_total_energy_expenditure,
        gpu_bar_plot,
    )

    log_dir_path = Path(log_dir)
    if not log_dir_path.exists():
        raise ValueError(
            f"The provided log_dir, '{log_dir_path.resolve()}', does not exist."
        )

    plot_dir_path = Path(plot_dir)
    if pattern.lower() == "none":
        pattern = None

    gpu_data_df = convert_matching_files_to_dataframe(
        pattern=pattern, log_dir=log_dir_path
    )

    energy_df = calculate_total_energy_expenditure(gpu_data_df=gpu_data_df)
    utilization_df = calculate_average_gpu_utilization(gpu_data_df=gpu_data_df)

    plot_dir_path.mkdir(parents=True, exist_ok=True)
    energy_plot_path = plot_dir_path / "gpu_energy_plot.png"
    utilization_plot_path = plot_dir_path / "utilization_plot.png"

    energy_fig, _ = gpu_bar_plot(
        data_df=energy_df,
        plot_title="Energy Consumption by Strategy and Number of GPUs",
        y_label="Energy Consumption (Wh)",
        main_column="total_energy_wh",
    )
    utilization_fig, _ = gpu_bar_plot(
        data_df=utilization_df,
        plot_title="GPU Utilization by Strategy and Number of GPUs",
        y_label="GPU Utilization (%)",
        main_column="utilization",
    )

    energy_fig.savefig(energy_plot_path)
    utilization_fig.savefig(utilization_plot_path)
    print(f"Saved GPU energy plot at '{energy_plot_path.resolve()}'.")
    print(f"Saved utilization plot at '{utilization_plot_path.resolve()}'.")

    if not do_backup:
        return

    backup_scalability_metrics(
        experiment_name=experiment_name,
        run_name=run_name,
        backup_dir=backup_dir,
        metric_df=gpu_data_df,
        filename="gpu_data.csv",
    )


@app.command()
def generate_communication_plot(
    log_dir: str = "scalability-metrics/communication-data",
    pattern: str = r"(.+)_(\d+)_(\d+)\.csv$",
    output_file: str = "plots/communication_plot.png",
    do_backup: bool = False,
    backup_dir: str = "backup-scalability-metrics/",
    experiment_name: Optional[str] = None,
    run_name: Optional[str] = None,
) -> None:
    """Generate stacked plot showing computation vs. communication fraction. Stores it
    to output_file.

    Args:
        log_dir: The directory where the csv logs are stored. Defaults to
            ``profiling_logs``.
        pattern: A regex pattern to recognize the file names in the 'log_dir' folder.
            Defaults to ``profile_(\\w+)_(\\d+)_(\\d+)\\.csv$``. Set it to 'None' to
            make it None. In this case, it will match all files in the given folder.
        output_file: The path to where the resulting plot should be saved. Defaults to
            ``plots/comm_plot.png``.
        do_backup: Whether to backup the data used for making the plot or not.
        backup_dir: The path to where the data used to produce the plot should be
            saved.
        experiment_name: The name of the experiment to be used when creating a backup
            of the data used for the plot.
        run_name: The name of the run to be used when creating a backup of the data
            used for the plot.
    """

    from itwinai.scalability import (
        backup_scalability_metrics,
        convert_matching_files_to_dataframe,
    )
    from itwinai.torch.profiling.communication_plot import (
        communication_overhead_stacked_bar_plot,
        get_comp_fraction_full_array,
    )

    log_dir_path = Path(log_dir)
    if not log_dir_path.exists():
        raise ValueError(
            f"The provided directory, '{log_dir_path.resolve()}', does not exist."
        )

    if pattern.lower() == "none":
        pattern = None

    expected_columns = {
        "strategy",
        "num_gpus",
        "global_rank",
        "name",
        "self_cuda_time_total",
    }
    communication_df = convert_matching_files_to_dataframe(
        log_dir=log_dir_path, pattern=pattern, expected_columns=expected_columns
    )
    values = get_comp_fraction_full_array(communication_df, print_table=True)

    strategies = sorted(communication_df["strategy"].unique())
    gpu_numbers = sorted(communication_df["num_gpus"].unique(), key=lambda x: int(x))

    fig, _ = communication_overhead_stacked_bar_plot(values, strategies, gpu_numbers)

    output_path = Path(output_file)
    output_path.parent.mkdir(parents=True, exist_ok=True)

    fig.savefig(output_path)
    print(f"\nSaved computation vs. communication plot at '{output_path.resolve()}'.")

    if not do_backup:
        return

    backup_scalability_metrics(
        experiment_name=experiment_name,
        run_name=run_name,
        backup_dir=backup_dir,
        metric_df=communication_df,
        filename="communication_data.csv",
    )


@app.command()
def generate_scalability_plot(
    pattern: str = "None",
    log_dir: str = "scalability-metrics/epoch-time",
    do_backup: bool = False,
    backup_dir: str = "backup-scalability-metrics/",
    experiment_name: Optional[str] = None,
    run_name: Optional[str] = None,
) -> None:
    """Creates two scalability plots from measured wall-clock times of an experiment
    run and saves them to file. Uses pattern to filter out files if given, otherwise
    it will try to use all files it finds in the given log directory. Will store all
    the data that was used as a backup file if do_backup is provided.
    """

    from itwinai.scalability import (  # archive_data,
        backup_scalability_metrics,
        convert_matching_files_to_dataframe,
        create_absolute_plot,
        create_relative_plot,
    )

    log_dir_path = Path(log_dir)
    if pattern.lower() == "none":
        pattern = None

    expected_columns = {"name", "nodes", "epoch_id", "time"}
    combined_df = convert_matching_files_to_dataframe(
        log_dir=log_dir_path, pattern=pattern, expected_columns=expected_columns
    )
    print("Merged CSV:")
    print(combined_df)

    avg_time_df = (
        combined_df.drop(columns="epoch_id")
        .groupby(["name", "nodes"])
        .mean()
        .reset_index()
    )
    print("\nAvg over name and nodes:")
    print(avg_time_df.rename(columns=dict(time="avg(time)")))

    create_absolute_plot(avg_time_df)
    create_relative_plot(avg_time_df)

    if not do_backup:
        return

    backup_scalability_metrics(
        experiment_name=experiment_name,
        run_name=run_name,
        backup_dir=backup_dir,
        metric_df=combined_df,
        filename="epoch_time.csv",
    )


@app.command()
def sanity_check(
    torch: Annotated[
        Optional[bool], typer.Option(help=("Check also itwinai.torch modules."))
    ] = False,
    tensorflow: Annotated[
        Optional[bool], typer.Option(help=("Check also itwinai.tensorflow modules."))
    ] = False,
    all: Annotated[Optional[bool], typer.Option(help=("Check all modules."))] = False,
    optional_deps: List[str] = typer.Option(
        None, help="List of optional dependencies."
    ),
):
    """Run sanity checks on the installation of itwinai and its dependencies by trying
    to import itwinai modules. By default, only itwinai core modules (neither torch, nor
    tensorflow) are tested."""
    from itwinai.tests.sanity_check import (
        run_sanity_check,
        sanity_check_all,
        sanity_check_slim,
        sanity_check_tensorflow,
        sanity_check_torch,
    )

    all = (torch and tensorflow) or all
    if all:
        sanity_check_all()
    elif torch:
        sanity_check_torch()
    elif tensorflow:
        sanity_check_tensorflow()
    else:
        sanity_check_slim()

    if optional_deps is not None:
        run_sanity_check(optional_deps)


@app.command()
def exec_pipeline(
    config: Annotated[
        Path,
        typer.Option(help="Path to the configuration file of the pipeline to execute."),
    ],
    pipe_key: Annotated[
        str,
        typer.Option(
            help=(
                "Key in the configuration file identifying "
                "the pipeline object to execute."
            )
        ),
    ] = "pipeline",
    steps: Annotated[
        Optional[str],
        typer.Option(
            help=(
                "Run only some steps of the pipeline. Accepted values are "
                "indices, python slices (e.g., 0:3 or 2:10:100), and "
                "string names of steps."
            )
        ),
    ] = None,
    print_config: Annotated[
        bool, typer.Option(help=("Print config to be executed after overrides."))
    ] = False,
    overrides_list: Annotated[
        Optional[List[str]],
        typer.Option(
            "--override",
            "-o",
            help=(
                "Nested key to dynamically override elements in the "
                "configuration file with the "
                "corresponding new value, joined by '='. It is also possible "
                "to index elements in lists using their list index. "
                "Example: [...] "
                "-o pipeline.init_args.trainer.init_args.lr=0.001 "
                "-o pipeline.my_list.2.batch_size=64 "
            ),
        ),
    ] = None,
):
    """Execute a pipeline from configuration file. Allows dynamic override of fields."""
    # Add working directory to python path so that the interpreter is able
    # to find the local python files imported from the pipeline file
    import os
    import re
    import sys

    from .utils import str_to_slice

    sys.path.append(os.path.dirname(config))
    sys.path.append(os.getcwd())

    # Parse and execute pipeline
    from itwinai.parser import ConfigParser

    overrides_list = overrides_list if overrides_list is not None else []
    overrides = {
        k: v
        for k, v in map(lambda x: (x.split("=")[0], x.split("=")[1]), overrides_list)
    }
    parser = ConfigParser(config=config, override_keys=overrides)
    if print_config:
        import json

        print()
        print("#=" * 15 + " Used configuration " + "#=" * 15)
        print(json.dumps(parser.config, indent=2))
        print("#=" * 50)
        print()

    pipeline = parser.parse_pipeline(pipeline_nested_key=pipe_key)
    if steps:
        if not re.match(r"\d+(:\d+)?(:\d+)?", steps):
            print(f"Looking for step name '{steps}'")
        else:
            steps = str_to_slice(steps)
        pipeline = pipeline[steps]
    pipeline.execute()


@app.command()
def mlflow_ui(
    path: str = typer.Option("ml-logs/", help="Path to logs storage."),
    port: int = typer.Option(5000, help="Port on which the MLFlow UI is listening."),
):
    """Visualize Mlflow logs."""
    import subprocess

    subprocess.run(f"mlflow ui --backend-store-uri {path} --port {port}".split())


@app.command()
def mlflow_server(
    path: str = typer.Option("ml-logs/", help="Path to logs storage."),
    port: int = typer.Option(5000, help="Port on which the server is listening."),
):
    """Spawn Mlflow server."""
    import subprocess

    subprocess.run(f"mlflow server --backend-store-uri {path} --port {port}".split())


@app.command()
def kill_mlflow_server(
    port: int = typer.Option(5000, help="Port on which the server is listening."),
):
    """Kill Mlflow server."""
    import subprocess

    subprocess.run(
        f"kill -9 $(lsof -t -i:{port})".split(), check=True, stderr=subprocess.DEVNULL
    )


if __name__ == "__main__":
    app()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Henry Mutegeki
#
# Credit:
# - Henry Mutegeki <henry.mutegeki@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import argparse
import os
from typing import Dict, Literal, Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
from torchvision import datasets, transforms

from itwinai.loggers import Logger, MLFlowLogger
from itwinai.torch.config import TrainingConfiguration
from itwinai.torch.distributed import DeepSpeedStrategy
from itwinai.torch.trainer import TorchTrainer
from itwinai.torch.type import Metric


class Generator(nn.Module):
    def __init__(self, z_dim, g_hidden, image_channel):
        super(Generator, self).__init__()
        self.main = nn.Sequential(
            # input layer
            nn.ConvTranspose2d(z_dim, g_hidden * 8, 4, 1, 0, bias=False),
            nn.BatchNorm2d(g_hidden * 8),
            nn.ReLU(True),
            # 1st hidden layer
            nn.ConvTranspose2d(g_hidden * 8, g_hidden * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(g_hidden * 4),
            nn.ReLU(True),
            # 2nd hidden layer
            nn.ConvTranspose2d(g_hidden * 4, g_hidden * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(g_hidden * 2),
            nn.ReLU(True),
            # 3rd hidden layer
            nn.ConvTranspose2d(g_hidden * 2, g_hidden, 4, 2, 1, bias=False),
            nn.BatchNorm2d(g_hidden),
            nn.ReLU(True),
            # output layer
            nn.ConvTranspose2d(g_hidden, image_channel, 4, 2, 1, bias=False),
            nn.Tanh(),
        )

    def forward(self, input):
        return self.main(input)


class Discriminator(nn.Module):
    def __init__(self, d_hidden, image_channel):
        super(Discriminator, self).__init__()
        self.main = nn.Sequential(
            # 1st layer
            nn.Conv2d(image_channel, d_hidden, 4, 2, 1, bias=False),
            nn.LeakyReLU(0.2, inplace=True),
            # 2nd layer
            nn.Conv2d(d_hidden, d_hidden * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(d_hidden * 2),
            nn.LeakyReLU(0.2, inplace=True),
            # 3rd layer
            nn.Conv2d(d_hidden * 2, d_hidden * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(d_hidden * 4),
            nn.LeakyReLU(0.2, inplace=True),
            # 4th layer
            nn.Conv2d(d_hidden * 4, d_hidden * 8, 4, 2, 1, bias=False),
            nn.BatchNorm2d(d_hidden * 8),
            nn.LeakyReLU(0.2, inplace=True),
            # output layer
            nn.Conv2d(d_hidden * 8, 1, 4, 1, 0, bias=False),
            nn.Sigmoid(),
        )

    def forward(self, input):
        return self.main(input).view(-1, 1).squeeze(1)


class GANTrainer(TorchTrainer):
    """Trainer class for GAN model using pytorch.

    Args:
        config (Union[Dict, TrainingConfiguration]): training configuration
            containing hyperparameters.
        epochs (int): number of training epochs.
        discriminator (nn.Module): pytorch discriminator model to train GAN.
        generator (nn.Module): pytorch generator model to train GAN.
        strategy (Literal['ddp', 'deepspeed', 'horovod'], optional):
            distributed strategy. Defaults to 'ddp'.
        validation_every (Optional[int], optional): run a validation epoch
            every ``validation_every`` epochs. Disabled if None. Defaults to 1.
        test_every (Optional[int], optional): run a test epoch
            every ``test_every`` epochs. Disabled if None. Defaults to None.
        random_seed (Optional[int], optional): set random seed for
            reproducibility. If None, the seed is not set. Defaults to None.
        logger (Optional[Logger], optional): logger for ML tracking.
            Defaults to None.
        metrics (Optional[Dict[str, Metric]], optional): map of torch metrics
            metrics. Defaults to None.
        checkpoints_location (str): path to checkpoints directory.
            Defaults to "checkpoints".
        checkpoint_every (Optional[int]): save a checkpoint every
            ``checkpoint_every`` epochs. Disabled if None. Defaults to None.
        name (Optional[str], optional): trainer custom name. Defaults to None.
    """

    def __init__(
        self,
        config: Union[Dict, TrainingConfiguration],
        epochs: int,
        discriminator: nn.Module,
        generator: nn.Module,
        strategy: Literal["ddp", "deepspeed"] = "ddp",
        validation_every: Optional[int] = 1,
        test_every: Optional[int] = None,
        random_seed: Optional[int] = None,
        logger: Optional[Logger] = None,
        metrics: Optional[Dict[str, Metric]] = None,
        checkpoints_location: str = "checkpoints",
        checkpoint_every: Optional[int] = None,
        name: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            config=config,
            epochs=epochs,
            model=None,
            strategy=strategy,
            validation_every=validation_every,
            test_every=test_every,
            random_seed=random_seed,
            logger=logger,
            metrics=metrics,
            checkpoints_location=checkpoints_location,
            checkpoint_every=checkpoint_every,
            name=name,
            **kwargs,
        )
        self.save_parameters(**self.locals2params(locals()))
        self.discriminator = discriminator
        self.generator = generator

    def create_model_loss_optimizer(self) -> None:
        self.optimizerD = optim.Adam(
            self.discriminator.parameters(), lr=self.config.lr, betas=(0.5, 0.999)
        )
        self.optimizerG = optim.Adam(
            self.generator.parameters(), lr=self.config.lr, betas=(0.5, 0.999)
        )
        self.criterion = nn.BCELoss()

        # https://stackoverflow.com/a/67437077
        self.discriminator = torch.nn.SyncBatchNorm.convert_sync_batchnorm(self.discriminator)
        self.generator = torch.nn.SyncBatchNorm.convert_sync_batchnorm(self.generator)

        # First, define strategy-wise optional configurations
        if isinstance(self.strategy, DeepSpeedStrategy):
            # Batch size definition is not optional for DeepSpeedStrategy!
            distribute_kwargs = dict(
                config_params=dict(train_micro_batch_size_per_gpu=self.config.batch_size)
            )
        else:
            distribute_kwargs = {}
        # Distribute discriminator and its optimizer
        self.discriminator, self.optimizerD, _ = self.strategy.distributed(
            self.discriminator, self.optimizerD, **distribute_kwargs
        )
        self.generator, self.optimizerG, _ = self.strategy.distributed(
            self.generator, self.optimizerG, **distribute_kwargs
        )
        self.discriminator.to(self.device)
        self.generator.to(self.device)

    def train_epoch(self, epoch: int):
        self.discriminator.train()
        self.generator.train()
        gen_train_losses = []
        disc_train_losses = []
        disc_train_accuracy = []
        for batch_idx, (real_images, _) in enumerate(self.train_dataloader):
            lossG, lossD, accuracy_disc = self.train_step(real_images, batch_idx)
            gen_train_losses.append(lossG)
            disc_train_losses.append(lossD)
            disc_train_accuracy.append(accuracy_disc)

            self.train_glob_step += 1
        # Aggregate and log losses and accuracy
        avg_disc_accuracy = torch.mean(torch.stack(disc_train_accuracy))
        self.log(
            item=avg_disc_accuracy.item(),
            identifier="disc_train_accuracy_per_epoch",
            kind="metric",
            step=epoch,
        )
        avg_gen_loss = torch.mean(torch.stack(gen_train_losses))
        self.log(
            item=avg_gen_loss.item(),
            identifier="gen_train_loss_per_epoch",
            kind="metric",
            step=epoch,
        )

        avg_disc_loss = torch.mean(torch.stack(disc_train_losses))
        self.log(
            item=avg_disc_loss.item(),
            identifier="disc_train_loss_per_epoch",
            kind="metric",
            step=epoch,
        )

        self.save_fake_generator_images(epoch)

    def validation_epoch(self, epoch: int):
        gen_validation_losses = []
        gen_validation_accuracy = []
        disc_validation_losses = []
        disc_validation_accuracy = []
        self.discriminator.eval()
        self.generator.eval()
        for batch_idx, (real_images, _) in enumerate(self.validation_dataloader):
            loss_gen, accuracy_gen, loss_disc, accuracy_disc = self.validation_step(
                real_images, batch_idx
            )
            gen_validation_losses.append(loss_gen)
            gen_validation_accuracy.append(accuracy_gen)
            disc_validation_losses.append(loss_disc)
            disc_validation_accuracy.append(accuracy_disc)
            self.validation_glob_step += 1

        # Aggregate and log metrics
        disc_validation_loss = torch.mean(torch.stack(disc_validation_losses))
        self.log(
            item=disc_validation_loss.item(),
            identifier="disc_valid_loss_per_epoch",
            kind="metric",
            step=epoch,
        )
        disc_validation_accuracy = torch.mean(torch.stack(disc_validation_accuracy))
        self.log(
            item=disc_validation_accuracy.item(),
            identifier="disc_valid_accuracy_epoch",
            kind="metric",
            step=epoch,
        )
        gen_validation_loss = torch.mean(torch.stack(gen_validation_losses))
        self.log(
            item=gen_validation_loss.item(),
            identifier="gen_valid_loss_per_epoch",
            kind="metric",
            step=epoch,
        )
        gen_validation_accuracy = torch.mean(torch.stack(gen_validation_accuracy))
        self.log(
            item=gen_validation_accuracy.item(),
            identifier="gen_valid_accuracy_epoch",
            kind="metric",
            step=epoch,
        )

        return gen_validation_loss

    def train_step(self, real_images, batch_idx):
        real_images = real_images.to(self.device)
        batch_size = real_images.size(0)
        real_labels = torch.ones((batch_size,), dtype=torch.float, device=self.device)
        fake_labels = torch.zeros((batch_size,), dtype=torch.float, device=self.device)

        # Train Discriminator with real images
        output_real = self.discriminator(real_images)
        lossD_real = self.criterion(output_real, real_labels)
        # Generate fake images and train Discriminator
        noise = torch.randn(batch_size, self.config.z_dim, 1, 1, device=self.device)

        fake_images = self.generator(noise)
        output_fake = self.discriminator(fake_images.detach())
        lossD_fake = self.criterion(output_fake, fake_labels)

        lossD = (lossD_real + lossD_fake) / 2

        self.optimizerD.zero_grad()
        lossD.backward()
        self.optimizerD.step()

        accuracy = ((output_real > 0.5).float() == real_labels).float().mean() + (
            (output_fake < 0.5).float() == fake_labels
        ).float().mean()
        accuracy_disc = accuracy.mean()

        # Train Generator
        output_fake = self.discriminator(fake_images)
        lossG = self.criterion(output_fake, real_labels)
        self.optimizerG.zero_grad()
        lossG.backward()
        self.optimizerG.step()
        self.log(
            item=accuracy_disc,
            identifier="disc_train_accuracy_per_batch",
            kind="metric",
            step=self.train_glob_step,
            batch_idx=batch_idx,
        )
        self.log(
            item=lossG,
            identifier="gen_train_loss_per_batch",
            kind="metric",
            step=self.train_glob_step,
            batch_idx=batch_idx,
        )
        self.log(
            item=lossD,
            identifier="disc_train_loss_per_batch",
            kind="metric",
            step=self.train_glob_step,
            batch_idx=batch_idx,
        )

        return lossG, lossD, accuracy_disc

    def validation_step(self, real_images, batch_idx):
        real_images = real_images.to(self.device)
        batch_size = real_images.size(0)
        real_labels = torch.ones((batch_size,), dtype=torch.float, device=self.device)
        fake_labels = torch.zeros((batch_size,), dtype=torch.float, device=self.device)

        # Validate with real images
        output_real = self.discriminator(real_images)
        loss_real = self.criterion(output_real, real_labels)

        # Generate and validate fake images
        noise = torch.randn(batch_size, self.config.z_dim, 1, 1, device=self.device)

        with torch.no_grad():
            fake_images = self.generator(noise)
            output_fake = self.discriminator(fake_images.detach())
        loss_fake = self.criterion(output_fake, fake_labels)

        # Generator's attempt to fool the discriminator
        loss_gen = self.criterion(output_fake, real_labels)
        accuracy_gen = ((output_fake > 0.5).float() == real_labels).float().mean()

        # Calculate total discriminator loss and accuracy
        loss_disc = (loss_real + loss_fake) / 2
        accuracy = ((output_real > 0.5).float() == real_labels).float().mean() + (
            (output_fake < 0.5).float() == fake_labels
        ).float().mean()
        accuracy_disc = accuracy.mean()

        self.log(
            item=loss_gen.item(),
            identifier="gen_valid_loss_per_batch",
            kind="metric",
            step=self.validation_glob_step,
            batch_idx=batch_idx,
        )
        self.log(
            item=accuracy_gen.item(),
            identifier="gen_valid_accuracy_per_batch",
            kind="metric",
            step=self.validation_glob_step,
            batch_idx=batch_idx,
        )

        self.log(
            item=loss_disc.item(),
            identifier="disc_valid_loss_per_batch",
            kind="metric",
            step=self.validation_glob_step,
            batch_idx=batch_idx,
        )
        self.log(
            item=accuracy_disc,
            identifier="disc_valid_accuracy_per_batch",
            kind="metric",
            step=self.validation_glob_step,
            batch_idx=batch_idx,
        )
        return loss_gen, accuracy_gen, loss_disc, accuracy_disc

    def save_checkpoint(self, name, epoch, loss=None):
        """Save training checkpoint with both optimizers."""
        if not os.path.exists(self.checkpoints_location):
            os.makedirs(self.checkpoints_location)

        checkpoint_path = os.path.join(self.checkpoints_location, f"{name}")
        checkpoint = {
            "epoch": epoch,
            "loss": loss.item() if loss is not None else None,
            "discriminator_state_dict": self.discriminator.state_dict(),
            "generator_state_dict": self.generator.state_dict(),
            "optimizerD_state_dict": self.optimizerD.state_dict(),
            "optimizerG_state_dict": self.optimizerG.state_dict(),
            "lr_scheduler": (self.lr_scheduler.state_dict() if self.lr_scheduler else None),
        }

        torch.save(checkpoint, checkpoint_path)
        print(f"Checkpoint saved to {checkpoint_path}")

    def load_checkpoint(self, checkpoint_path):
        """Load models and optimizers from checkpoint."""
        checkpoint = torch.load(checkpoint_path)

        self.discriminator.load_state_dict(checkpoint["discriminator_state_dict"])
        self.generator.load_state_dict(checkpoint["generator_state_dict"])
        self.optimizerD.load_state_dict(checkpoint["optimizerD_state_dict"])
        self.optimizerG.load_state_dict(checkpoint["optimizerG_state_dict"])

        if "lr_scheduler" in checkpoint and checkpoint["lr_scheduler"] is not None:
            self.lr_scheduler.load_state_dict(checkpoint["lr_scheduler"])

        print(f"Checkpoint loaded from {checkpoint_path}")

    def save_fake_generator_images(self, epoch):
        self.generator.eval()
        noise = torch.randn(64, self.config.z_dim, 1, 1, device=self.device)
        fake_images = self.generator(noise)
        fake_images_grid = torchvision.utils.make_grid(fake_images, normalize=True)
        fig, ax = plt.subplots(figsize=(8, 8))
        ax.set_axis_off()
        ax.set_title(f"Fake images for epoch {epoch}")
        ax.imshow(np.transpose(fake_images_grid.cpu().numpy(), (1, 2, 0)))
        self.log(
            item=fig,
            identifier=f"fake_images_epoch_{epoch}.png",
            kind="figure",
            step=epoch,
        )


def main():
    parser = argparse.ArgumentParser(description="PyTorch MNIST GAN Example")
    parser.add_argument(
        "--batch-size",
        type=int,
        default=128,
        help="input batch size for training (default: 128)",
    )
    parser.add_argument(
        "--epochs", type=int, default=15, help="number of epochs to train (default: 15)"
    )
    parser.add_argument(
        "--strategy", type=str, default="ddp", help="distributed strategy (default=ddp)"
    )
    parser.add_argument(
        "--lr", type=float, default=0.001, help="learning rate (default: 0.001)"
    )
    parser.add_argument("--seed", type=int, default=1, help="random seed (default: 1)")
    parser.add_argument(
        "--ckpt-interval",
        type=int,
        default=2,
        help="how many batches to wait before logging training status",
    )
    args = parser.parse_args()

    torch.manual_seed(args.seed)

    # Dataset
    transform = transforms.Compose(
        [
            transforms.Resize(64),
            transforms.ToTensor(),
            transforms.Normalize((0.5,), (0.5,)),
        ]
    )

    train_dataset = datasets.MNIST("../data", train=True, download=True, transform=transform)
    validation_dataset = datasets.MNIST("../data", train=False, transform=transform)

    def weights_init(m):
        classname = m.__class__.__name__
        if classname.find("Conv") != -1:
            m.weight.data.normal_(0.0, 0.02)
        elif classname.find("BatchNorm") != -1:
            m.weight.data.normal_(1.0, 0.02)
            m.bias.data.fill_(0)

    # Models
    netG = Generator(z_dim=100, g_hidden=64, image_channel=1)
    netG.apply(weights_init)
    netD = Discriminator(d_hidden=64, image_channel=1)
    netD.apply(weights_init)

    # Training configuration
    training_config = TrainingConfiguration(
        batch_size=args.batch_size, lr=args.lr, epochs=args.epochs, z_dim=100
    )

    # Logger
    logger = MLFlowLogger(experiment_name="Distributed GAN MNIST", log_freq=10)

    # Trainer
    trainer = GANTrainer(
        config=training_config,
        discriminator=netD,
        generator=netG,
        strategy=args.strategy,
        epochs=args.epochs,
        random_seed=args.seed,
        logger=logger,
    )

    # Launch training
    _, _, _, trained_model = trainer.execute(train_dataset, validation_dataset, None)


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------
from typing import Union

import torch.nn as nn
import torch.nn.functional as F
from torch import device
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader, Subset
from torchvision import datasets, transforms

from itwinai.parser import ArgumentParser as ItwinaiArgParser


def imagenet_dataset(data_root: str, subset_size: int | None = None):
    """Create a torch dataset object for Imagenet."""
    transform = transforms.Compose(
        [
            transforms.Resize(256),
            transforms.RandomHorizontalFlip(),
            transforms.RandomVerticalFlip(),
            transforms.RandomRotation(degrees=45),
            transforms.ColorJitter(
                brightness=0.5, contrast=0.5, saturation=0.5, hue=0.5
            ),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
        ]
    )
    imagenet = datasets.ImageFolder(root=data_root, transform=transform)

    if subset_size is None:
        # We do this because we always want to return an instance of a subset, to make
        # everything as consistent as possible
        subset_size = len(imagenet)
    if subset_size > len(imagenet):
        raise ValueError("Limit higher than the total length of the dataset")

    return Subset(imagenet, range(subset_size))


def train_epoch(
    model: nn.Module,
    device: device,
    train_loader: DataLoader,
    optimizer: Optimizer,
):
    """Train a pytorch model for a single epoch with the given arguments."""

    total_loss = 0
    model.train()

    for data, target in train_loader:
        data, target = data.to(device), target.to(device)
        optimizer.zero_grad()
        output = model(data)
        loss = F.nll_loss(output, target)
        loss.backward()
        optimizer.step()
        total_loss += loss.item()

    return total_loss


def get_parser() -> ItwinaiArgParser:
    parser = ItwinaiArgParser(description="PyTorch Imagenet scaling test")

    parser.add_argument(
        "--data-dir",
        default="./",
        help=("location of the training dataset in the " "local filesystem"),
    )
    parser.add_argument(
        "--log-int",
        type=int,
        default=10,
        help="log interval per training. Disabled if < 0.",
    )
    parser.add_argument(
        "--nworker",
        type=int,
        default=0,
        help=("number of workers in DataLoader " "(default: 0 - only main)"),
    )
    parser.add_argument(
        "--prefetch",
        type=int,
        default=2,
        help="prefetch data in DataLoader (default: 2)",
    )
    parser.add_argument(
        "--batch-size",
        type=int,
        default=64,
        help="input batch size for training (default: 64)",
    )
    parser.add_argument(
        "--epochs", type=int, default=10, help="number of epochs to train (default: 10)"
    )
    parser.add_argument(
        "--lr", type=float, default=0.01, help="learning rate (default: 0.01)"
    )
    parser.add_argument(
        "--momentum",
        type=float,
        default=0.5,
        help="momentum in SGD optimizer (default: 0.5)",
    )
    parser.add_argument(
        "--shuff",
        action="store_true",
        default=False,
        help="shuffle dataset (default: False)",
    )
    parser.add_argument(
        "--rnd-seed",
        type=int,
        default=None,
        help="seed integer for reproducibility (default: 0)",
    )
    parser.add_argument(
        "--backend",
        type=str,
        default="nccl",
        help="backend for parallelisation (default: nccl)",
    )
    parser.add_argument(
        "--subset-size",
        type=Union[int, None],
        default=None,
        help="How big of a subset of ImageNet to use during training.",
    )
    parser.add_argument(
        "--no-cuda", action="store_true", default=False, help="disables GPGPUs"
    )
    parser.add_argument(
        "--fp16-allreduce",
        action="store_true",
        default=False,
        help="use fp16 compression during allreduce",
    )
    parser.add_argument(
        "--use-adasum",
        action="store_true",
        default=False,
        help="use adasum algorithm to do reduction",
    )
    parser.add_argument(
        "--gradient-predivide-factor",
        type=float,
        default=1.0,
        help="apply gradient pre-divide factor in optimizer " "(default: 1.0)",
    )
    parser.add_argument(
        "--strategy",
        "-s",
        type=str,
        choices=["ddp", "horovod", "deepspeed"],
        default="ddp",
    )
    parser.add_argument(
        "--epoch-time-directory",
        type=str,
        default="scalability-metrics/epoch-time",
        help="Where to store the epoch time metrics used in the scalability report",
    )
    return parser



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------


from itwinai.slurm.slurm_script_builder import (
    SlurmScriptBuilder,
    SlurmScriptConfiguration,
)
from itwinai.slurm.utils import get_slurm_job_parser


class TutorialSlurmScriptBuilder(SlurmScriptBuilder):

    def __init__(
        self,
        slurm_script_configuration: SlurmScriptConfiguration,
        distributed_strategy: str,
        training_command: str | None = None,
        python_venv: str = ".venv",
        debug: bool = False,
        use_itwinai_trainer: bool = False,
    ):
        super().__init__(
            slurm_script_configuration=slurm_script_configuration,
            distributed_strategy=distributed_strategy,
            training_command=training_command,
            python_venv=python_venv,
            debug=debug,
        )
        self.use_itwinai_trainer = use_itwinai_trainer

    def generate_identifier(self) -> str:
        if self.use_itwinai_trainer:
            prepend = "itwinai-"
        else:
            prepend = "baseline-"
        return prepend + super().generate_identifier()

    def get_training_command(self):
        if self.training_command is not None:
            return self.training_command

        if self.use_itwinai_trainer:
            training_command = (
                "itwinai_trainer.py -c config/base.yaml "
                f"-c config/{self.distributed_strategy}.yaml -s {self.distributed_strategy}"
            )
        else:
            training_command = (
                f"{self.distributed_strategy}_trainer.py -c config/base.yaml "
                f"-c config/{self.distributed_strategy}.yaml"
            )

        if self.distributed_strategy == "horovod":
            training_command = "python " + training_command
        return training_command


def main():
    parser = get_slurm_job_parser()
    # Customizing the parser to this specific use case before retrieving the args
    parser.add_argument(
        "--itwinai-trainer",
        action="store_true",
        help="Whether to use the itwinai trainer or not.",
    )
    args = parser.parse_args()
    slurm_script_configuration = SlurmScriptConfiguration(
        job_name=args.job_name,
        account=args.account,
        time=args.time,
        partition=args.partition,
        std_out=args.std_out,
        err_out=args.err_out,
        num_nodes=args.num_nodes,
        num_tasks_per_node=args.num_tasks_per_node,
        gpus_per_node=args.gpus_per_node,
        cpus_per_gpu=args.cpus_per_gpu,
    )

    retain_file = not args.no_retain_file
    submit_job = not args.no_submit_job

    # Setting the training command for the single run
    if args.itwinai_trainer:
        training_command = (
            f"itwinai_trainer.py -c config/base.yaml "
            f"-c config/{args.dist_strat}.yaml -s {args.dist_strat}"
        )
    else:
        training_command = (
            f"{args.dist_strat}_trainer.py -c config/base.yaml "
            f"-c config/{args.dist_strat}.yaml"
        )
    if args.dist_strat == "horovod":
        training_command = "python " + training_command

    # Building the script
    script_builder = TutorialSlurmScriptBuilder(
        slurm_script_configuration=slurm_script_configuration,
        distributed_strategy=args.dist_strat,
        debug=args.debug,
        training_command=training_command,
        python_venv=args.python_venv,
        use_itwinai_trainer=args.itwinai_trainer,
    )

    # Processing the script depending on the given mode
    mode = args.mode
    if mode == "single":
        script_builder.process_slurm_script(
            retain_file=retain_file, submit_slurm_job=submit_job
        )
    elif mode == "runall":
        # Running all strategies with and without the itwinai trainer
        script_builder.training_command = None
        script_builder.use_itwinai_trainer = False
        script_builder.run_slurm_script_all_strategies(
            retain_file=retain_file, submit_slurm_job=submit_job
        )

        # We do this twice as there are two types of strategies
        script_builder.use_itwinai_trainer = True
        script_builder.run_slurm_script_all_strategies(
            retain_file=retain_file, submit_slurm_job=submit_job
        )
    elif mode == "scaling-test":
        # Running the scaling test with and without the itwinai trainer
        script_builder.training_command = None
        script_builder.use_itwinai_trainer = False
        script_builder.run_scaling_test(
            retain_file=retain_file,
            submit_slurm_job=submit_job,
            num_nodes_list=args.scalability_nodes,
        )

        script_builder.use_itwinai_trainer = True
        script_builder.run_scaling_test(
            retain_file=retain_file,
            submit_slurm_job=submit_job,
            num_nodes_list=args.scalability_nodes,
        )
    else:
        # This shouldn't really ever happen, but checking just in case
        raise ValueError(
            f"Mode can only be 'single', 'runall' or 'scaling-test', but was '{mode}'."
        )


if __name__ == "__main__":
    main()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

SLURM_TEMPLATE = r"""#!/bin/bash

# Job configuration
#SBATCH --job-name={job_name}
#SBATCH --account={account}
#SBATCH --partition={partition}
#SBATCH --time={time}

#SBATCH --output={std_out}
#SBATCH --error={err_out}

# Resources allocation
#SBATCH --nodes={num_nodes}
#SBATCH --ntasks-per-node={num_tasks_per_node}
#SBATCH --gpus-per-node={gpus_per_node}
#SBATCH --cpus-per-gpu={cpus_per_gpu}
#SBATCH --exclusive

{pre_exec_command}

{exec_command}"""

JUWELS_HPC_MODULES = [
    "Stages/2024",
    "GCC",
    "OpenMPI",
    "CUDA/12",
    "MPI-settings/CUDA",
    "Python/3.11.3",
    "HDF5",
    "PnetCDF",
    "libaio",
    "mpi4py",
]



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Jarl Sondre Sæther
#
# Credit:
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import subprocess
from pathlib import Path
from typing import List

from pydantic import BaseModel, Field

from itwinai.slurm.slurm_constants import JUWELS_HPC_MODULES, SLURM_TEMPLATE
from itwinai.slurm.utils import remove_indentation_from_multiline_string


class SlurmScriptConfiguration(BaseModel):
    """Configuration object for the SLURM script. It contains all the settings for the
    SLURM script such as which hardware you are requesting or for how long to run it.
    As it allows for any ``pre_exec_command`` and ``exec_command``, it should work for
    any SLURM script.
    """

    # Settings for the SLURM configuration
    job_name: str | None = Field(default=None)
    account: str
    partition: str
    time: str

    std_out: Path | None = Field(default=None)
    err_out: Path | None = Field(default=None)

    num_nodes: int
    num_tasks_per_node: int
    gpus_per_node: int
    cpus_per_gpu: int

    # Tyipcally used to set up the environment before executing the command,
    # e.g. "ml Python", "source .venv/bin/activate" etc.
    pre_exec_command: str | None = Field(default=None)

    # Command to execute, typically an 'srun' command
    exec_command: str | None = Field(default=None)

    def format_script(self) -> str:
        """Uses the provided configuration parameters and formats a SLURM script with
        the requested settings.

        Returns:
            str: A string containing the SLURM script.
        """
        return SLURM_TEMPLATE.format_map(self.model_dump())


class SlurmScriptBuilder:
    """Builds a SLURM script for various itwinai use cases, typically machine-learning
    use cases. Uses the provided ``SlurmScriptConfiguration`` to build the script. Some
    methods, such as the run-all or scaling-test methods, will modify the given
    configuration in-place, so be careful if reusing the configuration object.
    """

    def __init__(
        self,
        slurm_script_configuration: SlurmScriptConfiguration,
        distributed_strategy: str,
        training_command: str | None = None,
        python_venv: str = ".venv",
        debug: bool = False,
        file_folder: Path = Path("slurm_scripts"),
    ):
        self.slurm_script_configuration = slurm_script_configuration
        self.distributed_strategy = distributed_strategy
        self.training_command = training_command

        self.python_venv = python_venv
        self.debug = debug
        self.file_folder = file_folder

        if self.slurm_script_configuration.cpus_per_gpu > 0:
            self.omp_num_threads = self.slurm_script_configuration.cpus_per_gpu
        else:
            self.omp_num_threads = 1

    def generate_identifier(self) -> str:
        num_nodes = self.slurm_script_configuration.num_nodes
        gpus_per_node = self.slurm_script_configuration.gpus_per_node
        return f"{self.distributed_strategy}-{num_nodes}x{gpus_per_node}"

    def get_training_command(self) -> str:
        if self.training_command is None:
            raise ValueError(
                "self.training_command cannot be None in an instance of the parent class."
            )
        return self.training_command

    def get_pre_exec_command(self) -> str:
        """Generates a pre-execution command for the SLURM script. This will load the
        standard HPC modules, source the given python venv and set the OpenMP number
        of threads. Will also add debug echo statements if debug flag is set to True."""
        pre_exec_command = rf"""
            ml {" ".join(JUWELS_HPC_MODULES)}
            source {self.python_venv}/bin/activate
            export OMP_NUM_THREADS={self.omp_num_threads}
        """

        if self.debug:
            pre_exec_command += "\n" + self.get_debug_command()

        pre_exec_command = pre_exec_command.strip()
        return remove_indentation_from_multiline_string(pre_exec_command)

    def get_srun_command(self, torch_log_dir: str = "logs_torchrun") -> str:
        """Generates an srun command for the different distributed ML frameworks using
        the internal training command. Sets up rendezvous connections etc. when
        necessary and exports necessary environment variables.
        """
        if self.distributed_strategy in ["ddp", "deepspeed"]:
            rdzv_endpoint = (
                '\'$(scontrol show hostnames "$SLURM_JOB_NODELIST"'
                " | head -n 1)':29500"
            )
            main_command = rf"""
            srun --cpu-bind=none --ntasks-per-node=1 \
            bash -c "torchrun \
                --log_dir='{torch_log_dir}' \
                --nnodes=$SLURM_NNODES \
                --nproc_per_node=$SLURM_GPUS_PER_NODE \
                --rdzv_id=$SLURM_JOB_ID \
                --rdzv_conf=is_host=\$(((SLURM_NODEID)) && echo 0 || echo 1) \
                --rdzv_backend=c10d \
                --rdzv_endpoint={rdzv_endpoint} \
                {self.get_training_command()}"
        """
        else:
            gpus_per_node = self.slurm_script_configuration.gpus_per_node
            num_nodes = self.slurm_script_configuration.num_nodes
            num_tasks = gpus_per_node * num_nodes

            # E.g. "0,1,2,3" with 4 GPUs per node
            cuda_visible_devices = ",".join(str(i) for i in range(gpus_per_node))

            main_command = rf"""
            export CUDA_VISIBLE_DEVICES="{cuda_visible_devices}"
            srun --cpu-bind=none \
                --ntasks-per-node=$SLURM_GPUS_PER_NODE \
                --cpus-per-task=$SLURM_CPUS_PER_GPU \
                --ntasks={num_tasks} \
                {self.get_training_command()}
        """

        main_command = main_command.strip()
        return remove_indentation_from_multiline_string(main_command)

    def get_debug_command(self) -> str:
        """Generates a debug command to be added to the pre-exec command."""

        debug_print_command = rf"""
        echo "### SLURM Variables ###"
        echo "SLURM_SUBMIT_DIR: $SLURM_SUBMIT_DIR"
        echo "SLURM_JOB_ID: $SLURM_JOB_ID"
        echo "SLURM_JOB_NODELIST: $SLURM_JOB_NODELIST"
        echo "SLURM_NNODES: $SLURM_NNODES"
        echo "SLURM_NTASKS: $SLURM_NTASKS"
        echo "SLURM_TASKS_PER_NODE: $SLURM_TASKS_PER_NODE"
        echo "SLURM_SUBMIT_HOST: $SLURM_SUBMIT_HOST"
        echo "SLURMD_NODENAME: $SLURMD_NODENAME"
        echo "CUDA_VISIBLE_DEVICES: $CUDA_VISIBLE_DEVICES"
        echo ""
        echo "### Other Variables ###"
        echo "Distributed Strategy: {self.distributed_strategy}"
        echo "Current working directory: $(pwd)"
        echo "Which python: $(which python)"
        """
        debug_print_command = debug_print_command.strip()
        return remove_indentation_from_multiline_string(debug_print_command)

    def process_slurm_script(
        self,
        file_path: Path | None = None,
        retain_file: bool = True,
        submit_slurm_job: bool = True,
    ) -> None:
        """Will generate and process a SLURM script according to specifications. Will
        run the script if ``submit_slurm_job`` is set to True, will store the SLURM script file
        if ``retain_file`` is set to True. If both are false, it will simply print the
        script to the console without any further processing.

        Args:
            file_path: Where to store the file before processing the script. Also the location
                it will remain if ``retain_file`` is set to True.
            retain_file: Whether to keep or delete the file after finishing processing.
            submit_slurm_job: Whether to submit the script as a SLURM job.
        """
        self.slurm_script_configuration.pre_exec_command = self.get_pre_exec_command()
        self.slurm_script_configuration.exec_command = self.get_srun_command()

        # Setting some default fields
        if self.slurm_script_configuration.job_name is None:
            self.slurm_script_configuration.job_name = self.generate_identifier()

        if self.slurm_script_configuration.std_out is None:
            std_out_path = Path("slurm_job_logs") / (self.generate_identifier() + ".out")
            self.slurm_script_configuration.std_out = std_out_path
        if self.slurm_script_configuration.err_out is None:
            err_out_path = Path("slurm_job_logs") / (self.generate_identifier() + ".err")
            self.slurm_script_configuration.err_out = err_out_path

        # Making sure the std out and err out folders exist
        self.slurm_script_configuration.std_out.parent.mkdir(
            exist_ok=True, parents=True
        )
        self.slurm_script_configuration.err_out.parent.mkdir(
            exist_ok=True, parents=True
        )

        # Generate the script using the given configuration
        script = self.slurm_script_configuration.format_script()
        if not submit_slurm_job and not retain_file:
            print("#" * 20, "SLURM Script Preview", "#"*20)
            print(script)
            print("#" * 62)
            return

        if file_path is None:
            file_path = self.file_folder / (self.generate_identifier() + ".sh")

        if file_path.exists():
            raise ValueError(
                f"File '{file_path.resolve()}' already exists! Give a different path "
                f"or delete the file first!"
            )

        file_path.parent.mkdir(exist_ok=True, parents=True)
        with open(file_path, "w") as f:
            f.write(script)

        if submit_slurm_job:
            subprocess.run(["sbatch", str(file_path.resolve())])

        if not retain_file:
            file_path.unlink()

    def run_slurm_script_all_strategies(
        self,
        file_folder: Path = Path("slurm_scripts"),
        retain_file: bool = True,
        submit_slurm_job: bool = True,
        strategies: List[str] = ["ddp", "horovod", "deepspeed"],
    ):
        """Runs the SLURM script with all the given strategies. Does the same as the
        ``runall.sh`` script has been doing.
        """
        self.file_folder = file_folder
        for strategy in strategies:
            self.distributed_strategy = strategy

            # Overriding job_name, std_out and err_out
            self.slurm_script_configuration.job_name = self.generate_identifier()
            std_out_path = Path("slurm_job_logs") / (self.generate_identifier() + ".out")
            err_out_path = Path("slurm_job_logs") / (self.generate_identifier() + ".err")
            self.slurm_script_configuration.std_out = std_out_path
            self.slurm_script_configuration.err_out = err_out_path

            self.process_slurm_script(
                retain_file=retain_file, submit_slurm_job=submit_slurm_job
            )

    def run_scaling_test(
        self,
        file_folder: Path = Path("slurm_scripts"),
        retain_file: bool = True,
        submit_slurm_job: bool = True,
        strategies: List[str] = ["ddp", "horovod", "deepspeed"],
        num_nodes_list: List[int] = [1, 2, 4, 8],
    ):
        """Runs a scaling test, i.e. runs all the strategies with separate runs for each
        distinct number of nodes.
        """
        for num_nodes in num_nodes_list:
            self.slurm_script_configuration.num_nodes = num_nodes
            self.run_slurm_script_all_strategies(
                file_folder=file_folder,
                retain_file=retain_file,
                submit_slurm_job=submit_slurm_job,
                strategies=strategies,
            )



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Here we show how to implement component interfaces in a simple way.

New components are defined as classes inheriting from the BaseComponent class
or an already defined component.
"""

from typing import Any, List, Optional, Tuple

from itwinai.components import DataGetter, DataSplitter, Saver, Trainer, monitor_exec


class MyDataGetter(DataGetter):
    def __init__(self, data_size: int, name: Optional[str] = None) -> None:
        super().__init__(name)
        self.data_size = data_size
        self.save_parameters(data_size=data_size)

    @monitor_exec
    def execute(self) -> List[int]:
        """Return a list dataset.

        Returns:
            List[int]: dataset
        """
        return list(range(self.data_size))


class MyDatasetSplitter(DataSplitter):
    @monitor_exec
    def execute(self, dataset: List[int]) -> Tuple[List[int], List[int], List[int]]:
        """Splits a list dataset into train, validation and test datasets.

        Args:
            dataset (List[int]): input list dataset.

        Returns:
            Tuple[List[int], List[int], List[int]]: train, validation, and
            test datasets respectively.
        """
        train_n = int(len(dataset) * self.train_proportion)
        valid_n = int(len(dataset) * self.validation_proportion)
        train_set = dataset[:train_n]
        vaild_set = dataset[train_n : train_n + valid_n]
        test_set = dataset[train_n + valid_n :]
        return train_set, vaild_set, test_set


class MyTrainer(Trainer):
    def __init__(self, lr: float = 1e-3, name: Optional[str] = None) -> None:
        super().__init__(name)
        self.save_parameters(name=name, lr=lr)

    @monitor_exec
    def execute(
        self, train_set: List[int], vaild_set: List[int], test_set: List[int]
    ) -> Tuple[List[int], List[int], List[int], str]:
        """Dummy ML trainer mocking a ML training algorithm.

        Args:
            train_set (List[int]): training dataset.
            vaild_set (List[int]): validation dataset.
            test_set (List[int]): test dataset.

        Returns:
            Tuple[List[int], List[int], List[int], str]: train, validation,
            test datasets, and trained model.
        """
        return train_set, vaild_set, test_set, "my_trained_model"


class MySaver(Saver):
    @monitor_exec
    def execute(self, artifact: Any) -> Any:
        """Saves an artifact to disk.

        Args:
            artifact (Any): artifact to save (e.g., dataset, model).

        Returns:
            Any: input artifact.
        """
        return artifact



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# - Henry Mutegeki <henry.mutegeki@cern.ch> - CERN
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import abc
import functools
import os
from typing import Any, Callable, Iterable, List, Literal, Optional, Tuple, Union

import torch
import torch.distributed as dist
import torch.nn as nn
import torch.optim as optim
from torch.optim.lr_scheduler import _LRScheduler as LRScheduler
from torch.optim.optimizer import Optimizer
from torch.utils.data import DataLoader, Dataset, DistributedSampler, Sampler
from torch.utils.data.dataloader import T_co, _collate_fn_t, _worker_init_fn_t

from ..distributed import DistributedStrategy
from .type import DistributedStrategyError, UninitializedStrategyError


def distributed_resources_available() -> bool:
    """Check if the current execution environment
    has (enough) GPUs available to allow for distributed ML.

    Returns:
        bool: env can support distributed ML.
    """
    force_dist_env = int(os.environ.get("ITWINAI_FORCE_DIST", "0"))
    return force_dist_env or torch.cuda.is_available() and torch.cuda.device_count() > 1


def check_initialized(method: Callable) -> Callable:
    """Decorator for strategy methods to check whether the strategy
    was correctly initialized before calling the method."""

    @functools.wraps(method)
    def wrapper(self: "TorchDistributedStrategy", *args, **kwargs):
        if not self.is_initialized:
            raise UninitializedStrategyError(
                (
                    f"{self.__class__.__name__} has not been initialized. "
                    "Use the init method."
                )
            )
        return method(self, *args, **kwargs)

    return wrapper


def initialize_ray() -> None:
    """This method is used by the RayDDPStrategy and RayDeepSpeedStrategy to initialize
    the Ray backend if it is not already initialized.

        Raises:
            EnvironmentError: If required environment variables `HEAD_NODE_PORT` or
                `HEAD_NODE_IP` are not set.
                These should be set from the slurm script where the ray cluster is launched.
    """
    import ray

    if ray.is_initialized():
        return

    ray.init(address="auto")
    print(f"Nodes in the cluster: {ray.nodes()}")
    print(f"Available cluster resources: {ray.available_resources()}")


class TorchDistributedStrategy(DistributedStrategy):
    """Abstract class to define the distributed backend methods for
    PyTorch models.
    """

    #: Allows to discriminate distributed strategies from non-distributed.
    #: Defaults to True.
    is_distributed: bool = True
    #: Set to True when the current strategy is initialized.
    #: Defaults to False.
    is_initialized: bool = False

    # Provides the name of the strategy for logging purposes etc.
    name: str

    @property
    @check_initialized
    def is_main_worker(self) -> bool:
        """Checks if local worker has global rank equal to zero.

        Returns:
            bool: True if main worker.
        """
        return self.global_rank() == 0

    @abc.abstractmethod
    def init(self) -> None:
        """Initializes the chosen distributed backend"""

    @abc.abstractmethod
    def distributed(
        self,
        model: nn.Module,
        optimizer: Optimizer,
        lr_scheduler: Optional[LRScheduler] = None,
    ) -> Tuple[nn.Module, Optimizer, Optional[LRScheduler]]:
        """Setup model, optimizer and scheduler for distributed."""

    @abc.abstractmethod
    def global_world_size(self) -> int:
        """Returns the total number of processes (global world size).

        Returns:
            int: global world size.
        """

    @abc.abstractmethod
    def local_world_size(self) -> int:
        """Returns the number of local workers available on a node
        (local world size).
        Usually it is equal to the number of available GPUs.

        Returns:
            int: local world size.
        """

    @abc.abstractmethod
    def global_rank(self) -> int:
        """Returns the global rank of the current process.
        Rank ranges from 0 to world_size.

        Returns:
            int: global rank.
        """

    @abc.abstractmethod
    def local_rank(self) -> int:
        """Returns the local rank of the current process.

        Returns:
            int: local rank.
        """

    @check_initialized
    def device(self) -> str:
        """Device used by local worker.

        Returns:
            str: torch device in the form 'device:N' (e.g., 'cuda:0', 'cpu').
        """
        if torch.cuda.is_available():
            return f"cuda:{self.local_rank()}"
        return "cpu"

    def set_device(self):
        """Set local device."""
        if torch.cuda.is_available():
            torch.cuda.device(self.local_rank())
            # Needed by torch.distributed.gather_object
            torch.cuda.set_device(self.local_rank())

    @check_initialized
    def create_dataloader(
        self,
        dataset: Dataset[T_co],
        batch_size: Optional[int] = 1,
        shuffle: Optional[bool] = None,
        sampler: Union[Sampler, Iterable, None] = None,
        batch_sampler: Union[Sampler[List], Iterable[List], None] = None,
        num_workers: int = 0,
        collate_fn: Optional[_collate_fn_t] = None,
        pin_memory: bool = False,
        drop_last: bool = False,
        timeout: float = 0,
        worker_init_fn: Optional[_worker_init_fn_t] = None,
        multiprocessing_context=None,
        generator=None,
        *,
        prefetch_factor: Optional[int] = None,
        persistent_workers: bool = False,
        pin_memory_device: str = "",
    ):
        """Create a distributed DataLoader by using ``DistributedSampler`` as
        random sampler.

        Args:
            dataset (Dataset): dataset from which to load the data.
            batch_size (int, optional): how many samples per batch to load
                (default: ``1``).
            shuffle (bool, optional): set to ``True`` to have the data
                reshuffled at every epoch (default: ``False``).
            sampler (Sampler or Iterable, optional): defines the strategy to
                draw
                samples from the dataset. Can be any ``Iterable`` with
                ``__len__``
                implemented. If specified, :attr:`shuffle` must not be
                specified.
            batch_sampler (Sampler or Iterable, optional): like
                :attr:`sampler`, but
                returns a batch of indices at a time. Mutually exclusive with
                :attr:`batch_size`, :attr:`shuffle`, :attr:`sampler`,
                and :attr:`drop_last`.
            num_workers (int, optional): how many subprocesses to use for data
                loading. ``0`` means that the data will be loaded in the main
                process. (default: ``0``)
            collate_fn (Callable, optional): merges a list of samples to form a
                mini-batch of Tensor(s).  Used when using batched loading from
                a map-style dataset.
            pin_memory (bool, optional): If ``True``, the data loader will
                copy Tensors
                into device/CUDA pinned memory before returning them.  If your
                data elements
                are a custom type, or your :attr:`collate_fn` returns a batch
                that is a custom type,
                see the example below.
            drop_last (bool, optional): set to ``True`` to drop the last
                incomplete batch,
                if the dataset size is not divisible by the batch size.
                If ``False`` and
                the size of dataset is not divisible by the batch size, then
                the last batch
                will be smaller. (default: ``False``)
            timeout (numeric, optional): if positive, the timeout value for
                collecting a batch
                from workers. Should always be non-negative. (default: ``0``)
            worker_init_fn (Callable, optional): If not ``None``,
                this will be called on each
                worker subprocess with the worker id (an int in
                ``[0, num_workers - 1]``) as
                input, after seeding and before data loading.
                (default: ``None``)
            multiprocessing_context (str or
                multiprocessing.context.BaseContext, optional): If
                ``None``, the default `multiprocessing context`_ of
                your operating system will
                be used. (default: ``None``)
            generator (torch.Generator, optional): If not ``None``,
                this RNG will be used
                by RandomSampler to generate random indexes and
                multiprocessing to generate
                ``base_seed`` for workers. (default: ``None``)
            prefetch_factor (int, optional, keyword-only arg): Number of
                batches loaded
                in advance by each worker. ``2`` means there will be a total of
                2 * num_workers batches prefetched across all workers.
                (default value depends
                on the set value for num_workers. If value of num_workers=0
                default is ``None``.
                Otherwise, if value of ``num_workers > 0`` default is ``2``).
            persistent_workers (bool, optional): If ``True``, the data loader
                will not shut down
                the worker processes after a dataset has been consumed once.
                This allows to
                maintain the workers `Dataset` instances alive.
                (default: ``False``)
            pin_memory_device (str, optional): the device to
                :attr:`pin_memory` to if ``pin_memory`` is ``True``.

        Raises:
            UninitializedStrategyError: when this method is called for a
                strategy which had not been initialized.
            RuntimeError: when a user-provided sampler, if given, is not of
                type ``DistributedSampler``.

        .. warning:: If the ``spawn`` start method is used,
                    :attr:`worker_init_fn`
                    cannot be an unpicklable object, e.g., a lambda function.
                    See `Multiprocessing best practices`_ on more
                    details related to multiprocessing in PyTorch.


        .. warning:: ``len(dataloader)`` heuristic is based on the length of
                    the sampler used.
                    When :attr:`dataset` is an
                    :class:`~torch.utils.data.IterableDataset`,
                    it instead returns an estimate based on
                    ``len(dataset) / batch_size``, with proper
                    rounding depending on :attr:`drop_last`, regardless
                    of multi-process loading
                    configurations. This represents the best guess PyTorch
                    can make because PyTorch
                    trusts user :attr:`dataset` code in correctly handling
                    multi-process
                    loading to avoid duplicate data.

                    However, if sharding results in multiple workers having
                    incomplete last batches,
                    this estimate can still be inaccurate, because (1) an
                    otherwise complete batch can
                    be broken into multiple ones and (2) more than one batch
                    worth of samples can be
                    dropped when :attr:`drop_last` is set. Unfortunately,
                    PyTorch can not detect such cases in general.

                    See `Dataset Types`_ for more details on these two
                    types of datasets and how
                    :class:`~torch.utils.data.IterableDataset` interacts with
                    `Multi-process data loading`_.


        .. warning:: See `Reproducibility`_, and
                    `My data loader workers return identical random numbers`_,
                    and
                    `Randomness in multi-process data loading`_ notes for
                    random seed related questions.


        .. _multiprocessing context:
            https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
        .. _Multiprocessing best practices:
            https://pytorch.org/docs/stable/notes/multiprocessing.html#multiprocessing-best-practices
        .. _Reproducibility:
            https://pytorch.org/docs/stable/notes/randomness.html#reproducibility
        .. _My data loader workers return identical random numbers:
            https://pytorch.org/docs/stable/notes/faq.html#dataloader-workers-random-seed
        .. _Randomness in multi-process data loading:
            https://pytorch.org/docs/stable/data.html#data-loading-randomness
        .. _Multi-process data loading:
            https://pytorch.org/docs/stable/data.html#multi-process-data-loading
        .. _Dataset Types:
            https://pytorch.org/docs/stable/data.html#dataset-types
        """

        if batch_sampler is not None:
            print("WARNING: batch_sampler is ignored by TorchDistributedStrategy")

        if self.is_distributed:
            if sampler is None:
                sampler = DistributedSampler(
                    dataset,
                    num_replicas=self.global_world_size(),
                    rank=self.global_rank(),
                    shuffle=shuffle,
                )
            elif not isinstance(sampler, DistributedSampler):
                raise RuntimeError("User-provided sampler must implement DistributedSampler.")
        # shuffle and batch_sampler must be unset
        return DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            sampler=sampler,
            num_workers=num_workers,
            collate_fn=collate_fn,
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            worker_init_fn=worker_init_fn,
            multiprocessing_context=multiprocessing_context,
            generator=generator,
            prefetch_factor=prefetch_factor,
            persistent_workers=persistent_workers,
            pin_memory_device=pin_memory_device,
        )

    @abc.abstractmethod
    def clean_up(self) -> None:
        """Cleans up resources allocated by distributed strategy."""

    @abc.abstractmethod
    def allgather_obj(self, obj: Any) -> List[Any]:
        """All-gathers any object from the whole group in a list
        (to all workers).

        Args:
            obj (Any): object to gather from all workers.

        Returns:
            List[Any]: list of objects gathered from all workers.
        """

    @abc.abstractmethod
    def gather_obj(self, obj: Any, dst_rank: int = 0) -> List[Any]:
        """Gathers any object from the whole group in a list
        (to all workers).

        Args:
            obj (Any): object to gather from all workers.
            dst_rank (int): rank of the worker on which the objects list
                are gathered.

        Returns:
            List[Any]: list of objects gathered from all workers.
        """

    @abc.abstractmethod
    def gather(self, tensor: torch.Tensor, dst_rank: int = 0) -> Optional[List]:
        """Gathers any object from the whole group in a list
        (to all workers).

        Args:
            obj (Any): object to gather from all workers.
            dst_rank (int): rank of the worker on which the objects list
                are gathered.

        Returns:
            Optional[List[Any]]: list of objects gathered from all workers if main
                worker, otherwise return None.
        """


class TorchDDPStrategy(TorchDistributedStrategy):
    """PyTorch ``DistributedDataParallel`` distributed strategy class.

    Args:
        backend (Literal['nccl', 'gloo', 'mpi']): Name of the
            distributed communication backend to employ.
    """

    #: Torch distributed communication backend.
    backend: Literal["nccl", "gloo", "mpi"]

    def __init__(self, backend: Literal["nccl", "gloo", "mpi"]) -> None:
        super().__init__()
        self.backend = backend
        self.name = "torch-ddp"

    def init(self) -> None:
        """Initializes the distributed process group and the distributed
        package.

        Raises:
            RuntimeError: when there are not (enough) GPUs available.
            DistributedStrategyError: when trying to initialize a strategy
                which is already initialized.
        """
        if not distributed_resources_available():
            raise RuntimeError("Trying to run distributed on insufficient resources.")
        if self.is_initialized:
            raise DistributedStrategyError("Strategy was already initialized")
        dist.init_process_group(backend=self.backend)
        self.is_initialized = True

        self.set_device()

    @check_initialized
    def distributed(
        self,
        model: nn.Module,
        optimizer: Optimizer,
        lr_scheduler: Optional[LRScheduler] = None,
        find_unused_parameters: bool = False,
        **kwargs,
    ) -> Tuple[nn.Module, Optimizer, Optional[LRScheduler]]:
        """Setup model, optimizer and scheduler for distributed."""
        if torch.cuda.is_available():
            # If GPUs are available
            model = model.to(self.device())
            dist_model = torch.nn.parallel.DistributedDataParallel(
                model,
                device_ids=[self.device()],
                output_device=self.device(),
                find_unused_parameters=find_unused_parameters,
            )
        elif distributed_resources_available():
            # If GPUs are not available, but running distributed ML on CPUs
            dist_model = torch.nn.parallel.DistributedDataParallel(
                model,
                find_unused_parameters=find_unused_parameters,
            )
        else:
            dist_model = model

        return dist_model, optimizer, lr_scheduler

    @check_initialized
    def global_world_size(self) -> int:
        """Returns the total number of processes (global world size).

        Returns:
            int: global world size.
        """
        return dist.get_world_size()

    @check_initialized
    def local_world_size(self) -> int:
        """Returns the local number of workers available per node,
        which is usually the number of GPUs available.

        Returns:
            int: local world size.

        Raises:
            RuntimeError: when the local world size cannot be retrieved.
        """
        if torch.cuda.is_available():
            return torch.cuda.device_count()
        if "LOCAL_WORLD_SIZE" not in os.environ:
            raise RuntimeError(
                "Could not retrieve local world size as CUDA is unavailable and there is "
                "no 'LOCAL_WORLD_SIZE' environment variable."
            )
        return int(os.environ["LOCAL_WORLD_SIZE"])

    @check_initialized
    def global_rank(self) -> int:
        """Returns the global rank of the current process, where
        rank ranges from 0 to world_size.

        Returns:
            int: global rank.
        """
        return dist.get_rank()

    @check_initialized
    def local_rank(self) -> int:
        """Returns the local rank of the current process.

        Returns:
            int: local rank.
        """
        return dist.get_rank() % self.local_world_size()

    @check_initialized
    def clean_up(self) -> None:
        """Destroys the current process group."""
        if distributed_resources_available():
            dist.barrier()
            dist.destroy_process_group()

    @check_initialized
    def allgather_obj(self, obj: Any) -> List[Any]:
        """All-gathers any object from the whole group
        in a list (to all workers).

        Args:
            obj (Any): Object to gather from all workers.

        Returns:
            List[Any]: List of gathered objects.
        """
        # https://pytorch.org/docs/stable/distributed.html#collective-functions
        res = [None] * self.global_world_size()
        dist.all_gather_object(res, obj)
        return res

    @check_initialized
    def gather_obj(self, obj: Any, dst_rank: int = 0) -> Optional[List[Any]]:
        """Gathers any object from the whole group in a list
        (to all workers).

        Args:
            obj (Any): object to gather from all workers.
            dst_rank (int): rank of the worker on which the objects list
                are gathered.

        Returns:
            Optional[List[Any]]: list of objects gathered from all workers
            or ``None`` on non-destination ranks.
        """
        # https://pytorch.org/docs/stable/distributed.html#collective-functions
        if self.global_rank() == dst_rank:
            res = [None] * self.global_world_size()
            dist.gather_object(obj, res, dst=dst_rank)
            return res

        dist.gather_object(obj, dst=dst_rank)

    @check_initialized
    def gather(self, tensor: torch.Tensor, dst_rank: int = 0) -> Optional[List]:
        # https://pytorch.org/docs/stable/distributed.html#collective-functions

        # Ensure that the tensor is on the correct device (CUDA)
        tensor = tensor.to(self.device())
        if self.global_rank() != dst_rank:
            dist.gather(tensor, dst=dst_rank)
            return

        res = [
            torch.zeros_like(tensor, device=self.device())
            for _ in range(self.global_world_size())
        ]

        dist.gather(tensor, gather_list=res, dst=dst_rank)

        # Moving everything to the CPU before returning
        return [val.cpu() for val in res]


class DeepSpeedStrategy(TorchDistributedStrategy):
    """DeepSpeed distributed strategy class.

    Args:
        backend (Literal['nccl', 'gloo', 'mpi']): Name of the
            distributed communication backend to employ.
        config (Union[dict, Path, str]): DeepSpeed config. Either a
            dictionary or a path to a JSON file.
    """

    #: Torch distributed communication backend.
    backend: Literal["nccl", "gloo", "mpi"]

    def __init__(self, backend: Literal["nccl", "gloo", "mpi"]) -> None:
        super().__init__()
        self.backend = backend
        self.name = "deepspeed"

    def init(self) -> None:
        """Initializes the distributed process group and the distributed
        package.

        Raises:
            RuntimeError: when there are not (enough) GPUs available.
            DistributedStrategyError: when trying to initialize a strategy
                already initialized.
        """
        import deepspeed

        self.deepspeed = deepspeed
        if not distributed_resources_available():
            raise RuntimeError("Trying to run distributed on insufficient resources.")

        if self.is_initialized:
            raise DistributedStrategyError("Strategy was already initialized")

        # https://github.com/Lightning-AI/pytorch-lightning/issues/13567
        # This block of code should be removed as some point
        if os.environ.get("LOCAL_RANK"):
            os.environ["OMPI_COMM_WORLD_LOCAL_RANK"] = os.environ.get("LOCAL_RANK")

        # https://deepspeed.readthedocs.io/en/latest/initialize.html#training-initialization
        self.deepspeed.init_distributed(dist_backend=self.backend)
        self.is_initialized = True

        self.set_device()

    @check_initialized
    def distributed(
        self,
        model: nn.Module,
        optimizer: Optional[Optimizer] = None,
        lr_scheduler: Optional[LRScheduler] = None,
        model_parameters: Optional[Any] = None,
        **init_kwargs,
    ) -> Tuple[nn.Module, Optimizer, Optional[LRScheduler]]:
        """Setup model, optimizer and scheduler for distributed."""

        distrib_model, optimizer, _, lr_scheduler = self.deepspeed.initialize(
            model=model,
            model_parameters=model_parameters,
            optimizer=optimizer,
            lr_scheduler=lr_scheduler,
            dist_init_required=True,
            **init_kwargs,
        )
        return distrib_model, optimizer, lr_scheduler

    @check_initialized
    def global_world_size(self) -> int:
        """Returns the total number of processes (global world size).

        Returns:
            int: global world size.
        """
        return dist.get_world_size()

    @check_initialized
    def local_world_size(self) -> int:
        """Returns the local number of workers available per node,
        which is usually the number of GPUs available.

        Returns:
            int: local world size.

        Raises:
            RuntimeError: when the local world size cannot be retrieved.
        """
        if torch.cuda.is_available():
            return torch.cuda.device_count()
        if "LOCAL_WORLD_SIZE" not in os.environ:
            raise RuntimeError(
                "Could not retrieve local world size as CUDA is unavailable and there is "
                "no 'LOCAL_WORLD_SIZE' environment variable."
            )
        return int(os.environ["LOCAL_WORLD_SIZE"])

    @check_initialized
    def global_rank(self) -> int:
        """Returns the global rank of the current process, where
        rank ranges from 0 to world_size.

        Returns:
            int: global rank.
        """
        return dist.get_rank()

    @check_initialized
    def local_rank(self) -> int:
        """Returns the local rank of the current process.

        Returns:
            int: local rank.
        """
        return dist.get_rank() % self.local_world_size()

    @check_initialized
    def clean_up(self) -> None:
        """Destroys the current process group."""
        # deepspeed.sys.exit() # disabled as it kills the execution

    @check_initialized
    def allgather_obj(self, obj: Any) -> List[Any]:
        """All-gathers any object from the whole group
        in a list (to all workers).

        Args:
            obj (Any): Object to gather from all workers.

        Returns:
            List[Any]: List of gathered objects.
        """
        # https://pytorch.org/docs/stable/distributed.html#collective-functions
        res = [None] * self.global_world_size()
        dist.all_gather_object(res, obj)
        return res

    @check_initialized
    def gather_obj(self, obj: Any, dst_rank: int = 0) -> Optional[List[Any]]:
        """Gathers any object from the whole group in a list
        (to all workers).

        Args:
            obj (Any): object to gather from all workers.
            dst_rank (int): rank of the worker on which the objects list
                are gathered.

        Returns:
            Optional[List[Any]]: list of objects gathered from all workers
            or ``None`` on non-destination ranks.
        """
        # https://pytorch.org/docs/stable/distributed.html#collective-functions
        if self.global_rank() == dst_rank:
            res = [None] * self.global_world_size()
            dist.gather_object(obj, res, dst=dst_rank)
            return res

        dist.gather_object(obj, dst=dst_rank)

    @check_initialized
    def gather(self, tensor: torch.Tensor, dst_rank: int = 0) -> Optional[List[torch.Tensor]]:
        """Gathers a tensor from the whole group in a list
        (to all workers).

        Args:
            obj (Any): object to gather from all workers.
            dst_rank (int): rank of the worker on which the objects list
                are gathered.

        Returns:
            Optional[List[torch.Tensor]]: list of tensors gathered from all workers
            or ``None`` on non-destination ranks.
        """
        # https://pytorch.org/docs/stable/distributed.html#collective-functions

        # Ensure that the tensor is on the correct device (CUDA)
        tensor = tensor.to(self.device())
        if self.global_rank() != dst_rank:
            dist.gather(tensor, dst=dst_rank)
            return

        res = [
            torch.zeros_like(tensor, device=self.device())
            for _ in range(self.global_world_size())
        ]

        dist.gather(tensor, gather_list=res, dst=dst_rank)

        # Moving all the tensors to CPU before returning
        return [val.cpu() for val in res]


class HorovodStrategy(TorchDistributedStrategy):
    """Horovod distributed strategy class."""

    def __init__(self):
        super().__init__()
        self.name = "horovod"

    def init(self) -> None:
        """Initializes the Horovod distributed backend.

        Raises:
            RuntimeError: when there are not (enough) GPUs available.
            DistributedStrategyError: when trying to initialize a strategy
                already initialized.
        """
        if not distributed_resources_available():
            raise RuntimeError("Trying to run distributed on insufficient resources.")
        if self.is_initialized:
            raise DistributedStrategyError("Strategy was already initialized")

        import horovod.torch as hvd

        self.hvd = hvd

        self.hvd.init()
        self.is_initialized = True

        self.set_device()

    @check_initialized
    def distributed(
        self,
        model: nn.Module,
        optimizer: Optional[Optimizer] = None,
        lr_scheduler: Optional[LRScheduler] = None,
        **optim_kwargs,
    ) -> Tuple[nn.Module, Optimizer, Optional[LRScheduler]]:
        """Setup model, optimizer and scheduler for distributed."""

        model.to(self.device())

        # Scale learning rate
        # https://github.com/horovod/horovod/issues/1653#issuecomment-574764452
        lr_scaler = 1
        if optim_kwargs.get("op") == self.hvd.Adasum:
            lr_scaler = self.hvd.local_size()
        elif optim_kwargs.get("op") == self.hvd.Average:
            lr_scaler = self.hvd.size()
        for g in optimizer.param_groups:
            g["lr"] *= lr_scaler

        self._broadcast_params(model, optimizer)

        distOptimizer = self.hvd.DistributedOptimizer(
            optimizer, named_parameters=model.named_parameters(), **optim_kwargs
        )
        return model, distOptimizer, lr_scheduler

    def _broadcast_params(self, model: nn.Module, optimizer: optim.Optimizer) -> None:
        """Broadcasts variables from root rank to all other processes.

        Args:
            model (nn.Module): ML model that is to be broadcasted
                across processes.
            optimizer (optim.Optimizer): Optimizer that is to be broadcasted
                across processes.
        """
        self.hvd.broadcast_parameters(model.state_dict(), root_rank=0)
        self.hvd.broadcast_optimizer_state(optimizer, root_rank=-0)

    @check_initialized
    def global_world_size(self) -> int:
        """Returns the total number of processes (global world size).

        Returns:
            int: global world size.
        """
        return self.hvd.size()

    @check_initialized
    def local_world_size(self) -> int:
        """Returns the local number of workers available per node,
        which is usually the number of GPUs available.

        Returns:
            int: local world size.
        """
        return self.hvd.local_size()

    @check_initialized
    def global_rank(self) -> int:
        """Returns the global rank of the current process, where
        rank ranges from 0 to world_size.

        Returns:
            int: global rank.
        """
        return self.hvd.rank()

    @check_initialized
    def local_rank(self) -> int:
        """Returns the local rank of the current process.

        Returns:
            int: local rank.
        """
        return self.hvd.local_rank()

    @check_initialized
    def clean_up(self) -> None:
        """Shuts Horovod down."""
        self.hvd.shutdown()

    @check_initialized
    def allgather_obj(self, obj: Any) -> list[Any]:
        """All-gathers any object from the whole group
        in a list (to all workers).

        Args:
            obj (Any): Object to gather from all workers.

        Returns:
            List[Any]: List of gathered objects.
        """
        return self.hvd.allgather_object(obj)

    @check_initialized
    def gather_obj(self, obj: Any, dst_rank: int = 0) -> Optional[list[Any]]:
        """Gathers any object from the whole group in a list
        (to all workers). Under the hood it relies on allgather as gather is
        not supported by Horovod.

        Args:
            obj (Any): object to gather from all workers.
            dst_rank (int): rank of the worker on which the objects list
                are gathered.

        Returns:
            Optional[List[Any]]: list of objects gathered from all workers
            or ``None`` on non-destination ranks.
        """
        result = self.allgather_obj(obj)
        if self.global_rank() == dst_rank:
            # Return only if on rank == dst_rank
            return result

    @check_initialized
    def gather(self, tensor: torch.Tensor, dst_rank: int = 0) -> Optional[List[torch.Tensor]]:
        """Gathers a tensor from the whole group in a list
        (to all workers). Under the hood it relies on allgather as gather is
        not supported by Horovod.

        Args:
            obj (Any): object to gather from all workers.
            dst_rank (int): rank of the worker on which the objects list
                are gathered.

        Returns:
            Optional[List[torch.Tensor]]: list of tensors gathered from all workers
            or ``None`` on non-destination ranks.
        """
        result = self.allgather_obj(tensor)
        if self.global_rank() == dst_rank:
            # Return only if on rank == dst_rank
            # Moving all the tensors to CPU before returning
            return [val.cpu() for val in result]


class NonDistributedStrategy(TorchDistributedStrategy):
    """Dummy class for non-distributed environments."""

    #: This strategy is not distributed.
    #: Defaults to False.
    is_distributed: bool = True
    is_distributed: bool = False

    def __init__(self):
        super().__init__()
        self.name = "non-distributed"

    def init(self) -> None:
        """If CUDA is available set CUDA device, and do nothing more.

        Raises:
            DistributedStrategyError: when trying to initialize a strategy
                already initialized.
        """
        if self.is_initialized:
            raise DistributedStrategyError("Strategy was already initialized")
        self.set_device()
        self.is_initialized = True

    @check_initialized
    def distributed(
        self,
        model: nn.Module,
        optimizer: Optional[Optimizer] = None,
        lr_scheduler: Optional[LRScheduler] = None,
        **kwargs,
    ) -> Tuple[nn.Module, Optimizer, Optional[LRScheduler]]:
        """Do nothing and return model, optimizer and scheduler."""
        if torch.cuda.is_available():
            model = model.cuda()
        return model, optimizer, lr_scheduler

    def global_world_size(self) -> int:
        """Returns the total number of processes (global world size).

        Returns:
            int: global world size.
        """
        return 1

    def local_world_size(self) -> int:
        """Returns the local number of workers available per node,
        which is usually the number of GPUs available.

        Returns:
            int: local world size.
        """
        return 1

    def global_rank(self) -> int:
        """Returns the global rank of the current process, where
        rank ranges from 0 to world_size.

        Returns:
            int: global rank.
        """
        return 0

    def local_rank(self) -> int:
        """Returns the local rank of the current process.

        Returns:
            int: local rank.
        """
        return 0

    def clean_up(self) -> None:
        """Do nothing."""

    def allgather_obj(self, obj: Any) -> list[Any]:
        """Wraps ``obj`` into a List object.

        Args:
            obj (Any): object in a worker.

        Returns:
            list[Any]: input object wrapped in a list.
        """
        return [obj]

    def gather_obj(self, obj: Any, dst_rank: int = 0) -> list[Any]:
        """Wraps ``obj`` into a List object.

        Args:
            obj (Any): object in a worker.
            dst_rank (int): ignored.

        Returns:
            list[Any]: input object wrapped in a list.
        """
        return [obj]

    def gather(self, tensor: torch.Tensor, dst_rank: int = 0):
        """Wraps ``tensor`` into a List object.

        Args:
            tensor (Any): object in a worker.
            dst_rank (int): ignored.

        Returns:
            list[Any]: input object wrapped in a list.
        """
        return [tensor]


class RayDDPStrategy(TorchDDPStrategy):
    """A distributed data-parallel (DDP) strategy using Ray Train for PyTorch training."""

    def __init__(self) -> None:
        initialize_ray()
        import ray.train

        self.ray_train = ray.train

    def init(self) -> None:
        self.is_initialized = True

    @check_initialized
    def global_world_size(self) -> int:
        return self.ray_train.get_context().get_world_size()

    @check_initialized
    def local_world_size(self) -> int:
        return self.ray_train.get_context().get_local_world_size()

    @check_initialized
    def global_rank(self) -> int:
        return self.ray_train.get_context().get_world_rank()

    @check_initialized
    def local_rank(self) -> int:
        return self.ray_train.get_context().get_local_rank()

    @check_initialized
    def distributed(
        self,
        model: nn.Module,
        optimizer: Optimizer,
        lr_scheduler: Optional[LRScheduler] = None,
    ) -> Tuple[nn.Module, Optimizer, LRScheduler | None]:
        model = self.ray_train.torch.prepare_model(model)

        return model, optimizer, lr_scheduler

    @check_initialized
    def create_dataloader(
        self,
        dataset: Dataset[T_co],
        batch_size: Optional[int] = 1,
        shuffle: Optional[bool] = None,
        sampler: Union[Sampler, Iterable, None] = None,
        batch_sampler: Union[Sampler[List], Iterable[List], None] = None,
        num_workers: int = 0,
        collate_fn: Optional[_collate_fn_t] = None,
        pin_memory: bool = False,
        drop_last: bool = False,
        timeout: float = 0,
        worker_init_fn: Optional[_worker_init_fn_t] = None,
        multiprocessing_context=None,
        generator=None,
        *,
        prefetch_factor: Optional[int] = None,
        persistent_workers: bool = False,
        pin_memory_device: str = "",
    ):
        dataloader = DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            sampler=sampler,
            shuffle=shuffle,
            num_workers=num_workers,
            collate_fn=collate_fn,
            pin_memory=pin_memory,
            drop_last=drop_last,
            timeout=timeout,
            worker_init_fn=worker_init_fn,
            multiprocessing_context=multiprocessing_context,
            generator=generator,
            prefetch_factor=prefetch_factor,
            persistent_workers=persistent_workers,
            pin_memory_device=pin_memory_device,
        )

        return self.ray_train.torch.prepare_data_loader(dataloader)


class RayDeepSpeedStrategy(DeepSpeedStrategy):
    """A distributed strategy using Ray and DeepSpeed for PyTorch training.

    Args:
        backend (Literal["nccl", "gloo", "mpi"]): The backend for distributed communication.
    """

    def __init__(self, backend: Literal["nccl", "gloo", "mpi"]) -> None:
        initialize_ray()
        super().__init__(backend=backend)



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Anna Lappe
#
# Credit:
# - Anna Lappe <anna.elisa.lappe@cern.ch> - CERN
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import argparse
import sys
from pathlib import Path
from typing import Tuple

from torch.utils.data import Dataset, random_split
from torchvision import datasets, transforms

from itwinai.components import DataGetter, DataSplitter

data_dir = Path("data")


def download_fashion_mnist() -> None:
    """Download the FashionMNIST dataset using torchvision."""
    print("Downloading FashionMNIST dataset...")
    datasets.FashionMNIST(
        data_dir,
        train=True,
        download=True,
        transform=transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))]
        ),
    )
    datasets.FashionMNIST(
        data_dir,
        train=False,
        download=True,
        transform=transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))]
        ),
    )
    print("Download complete!")


class FashionMNISTGetter(DataGetter):
    def __init__(self) -> None:
        super().__init__()

    def execute(self) -> Tuple[Dataset, Dataset]:
        """Load the FashionMNIST dataset from the specified directory."""
        print("Loading FashionMNIST dataset...")
        train_dataset = datasets.FashionMNIST(
            data_dir,
            train=True,
            download=False,
            transform=transforms.Compose(
                [transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))]
            ),
        )
        print("Loading complete!")
        return train_dataset


class FashionMNISTSplitter(DataSplitter):
    def __init__(
        self,
        train_proportion: float,
        validation_proportion: float,
        test_proportion: float,
        name: str | None = None,
    ) -> None:
        super().__init__(train_proportion, validation_proportion, test_proportion, name)

    def execute(self, dataset: Dataset) -> Tuple[Dataset, Dataset, Dataset]:
        """Split the dataset into train, validation, and test sets."""
        print("Splitting dataset...")
        total_size = len(dataset)
        train_size = int(self.train_proportion * total_size)
        validation_size = int(self.validation_proportion * total_size)
        test_size = total_size - train_size - validation_size

        train_dataset, validation_dataset, test_dataset = random_split(
            dataset, [train_size, validation_size, test_size]
        )
        print("Splitting complete!")
        return train_dataset, validation_dataset, test_dataset


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="FashionMNIST Dataset Loader")
    parser.add_argument(
        "--download_only",
        action="store_true",
        help="Download the FashionMNIST dataset and exit",
    )
    args = parser.parse_args()

    if args.download_only:
        download_fashion_mnist()
        sys.exit()



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# - Anna Elisa Lappe <anna.elisa.lappe@cern.ch> - CERN
# - Jarl Sondre Sæther <jarl.sondre.saether@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import os
from pathlib import Path
from typing import Optional, Tuple

import h5py
import pandas as pd
import torch
from sklearn.model_selection import train_test_split
from torch.utils.data import Dataset, TensorDataset, random_split

from itwinai.components import DataGetter, DataProcessor, DataSplitter, monitor_exec
from src.dataset import (
    generate_cut_image_dataset,
    generate_dataset_aux_channels,
    generate_dataset_main_channel,
    normalize_,
)


class TimeSeriesDatasetGenerator(DataGetter):
    def __init__(self, data_root: str = "data", name: Optional[str] = None) -> None:
        """Initialize the TimeSeriesDatasetGenerator class.

        Args:
            data_root (str): Root folder where datasets will be saved.
            name (Optional[str]): Name of the data getter component.
        """
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))
        self.data_root = data_root
        if not os.path.exists(data_root):
            os.makedirs(data_root, exist_ok=True)

    @monitor_exec
    def execute(self) -> pd.DataFrame:
        """Generate a time-series dataset, convert it to Q-plots,
        save it to disk, and return it.

        Returns:
            pd.DataFrame: dataset of Q-plot images.
        """
        df_aux_ts = generate_dataset_aux_channels(
            1000,
            3,
            duration=16,
            sample_rate=500,
            num_waves_range=(20, 25),
            noise_amplitude=0.6,
        )
        df_main_ts = generate_dataset_main_channel(
            df_aux_ts, weights=None, noise_amplitude=0.1
        )

        # save datasets
        save_name_main = "TimeSeries_dataset_synthetic_main.pkl"
        save_name_aux = "TimeSeries_dataset_synthetic_aux.pkl"
        df_main_ts.to_pickle(os.path.join(self.data_root, save_name_main))
        df_aux_ts.to_pickle(os.path.join(self.data_root, save_name_aux))

        # Transform to images and save to disk
        df_ts = pd.concat([df_main_ts, df_aux_ts], axis=1)
        df = generate_cut_image_dataset(
            df_ts, list(df_ts.columns), num_processes=20, square_size=64
        )
        save_name = "Image_dataset_synthetic_64x64.pkl"
        df.to_pickle(os.path.join(self.data_root, save_name))
        return df


class SyntheticTimeSeriesDatasetHDF5(Dataset):
    def __init__(
        self,
        hdf5_file_location: str,
        chunk_size: int = 500,
        hdf5_dataset_name: str = "virgo_dataset",
    ):
        """Initialize the DataFrameDataset class.

        Args:
            hdf5_file_location: Location of the HDF5 file containing the dataset
            chunk_size: How many rows of data points each sample should contain
            hdf5_dataset_name: The name of the Dataset object of the HDF5 file
        """
        file_path = Path(hdf5_file_location)
        if not file_path.exists():
            raise ValueError(f"Given file location, {file_path.resolve()} does not exist. ")
        self.hdf5_dataset_name = hdf5_dataset_name
        self.file_path = file_path
        self.chunk_size = chunk_size
        with h5py.File(self.file_path, "r") as f:
            if self.hdf5_dataset_name not in f:
                raise ValueError(
                    f"The dataset {self.hdf5_dataset_name} does not exist in "
                    f"the given HDF5 file."
                )
            dataset = f[self.hdf5_dataset_name]
            self.num_datapoints = dataset.shape[0]

        if self.num_datapoints == 0:
            raise ValueError("The given file contains 0 data points!")

        if self.num_datapoints % self.chunk_size != 0:
            print(
                f"[WARNING]: Number of datapoints, {self.num_datapoints} is not "
                f"divisible by the given chunk size, {self.chunk_size}, so the "
                f"remainder will be truncated. "
            )
        self.length = self.num_datapoints // self.chunk_size
        print(f"self.length: {self.length}")

    def __len__(self):
        """Return the total number of files in the dataset."""
        return self.length

    def __getitem__(self, idx) -> torch.Tensor:
        """Retrieve a data sample by index and normalize.

        Args:
            idx (int): Index of the file to retrieve.

        Returns:
            torch.Tensor: Normalized tensor for specific idx
        """
        if idx >= len(self):
            raise ValueError(f"Index {idx} out of bounds for dataset with length {len(self)}!")

        offset = idx * self.chunk_size
        with h5py.File(self.file_path, "r") as f:
            dataset = f[self.hdf5_dataset_name]
            data = dataset[offset : offset + self.chunk_size]

        data = torch.tensor(data, dtype=torch.float32)
        return normalize_(data)


class TimeSeriesDatasetSplitter(DataSplitter):
    def __init__(
        self,
        train_proportion: int | float,
        validation_proportion: int | float = 0.0,
        test_proportion: int | float = 0.0,
        rnd_seed: Optional[int] = None,
        name: Optional[str] = None,
        hdf5_file_location: str = "data/virgo_data.hdf5",
        hdf5_dataset_name: str = "virgo_dataset",
        chunk_size: int = 500,
    ) -> None:
        """Initialize the splitter for time-series datasets.

        Args:
            train_proportion (int | float): Proportion of files for the training set.
            validation_proportion (int | float): Proportion for validation.
            test_proportion (int | float): Proportion for testing.
            rnd_seed (int | None): Seed for randomization (optional).
            root_folder (str | None): Folder containing the dataset files.
            name (str | None): Name of the data splitter.
        """
        super().__init__(train_proportion, validation_proportion, test_proportion, name)
        self.save_parameters(**self.locals2params(locals()))
        self.rnd_seed = rnd_seed
        self.hdf5_file_location = hdf5_file_location
        self.hdf5_dataset_name = hdf5_dataset_name
        self.chunk_size = chunk_size

    @monitor_exec
    def execute(self) -> Tuple[Dataset, Dataset, Dataset]:
        """Execute the dataset splitting process.

        Finds all pickled files in the root folder, then splits them into
        training, validation, and test sets based on the specified proportions.

        Returns:
            Tuple[Dataset, Dataset, Dataset]: Training, validation, and test datasets.
        """

        whole_dataset = SyntheticTimeSeriesDatasetHDF5(
            hdf5_file_location=self.hdf5_file_location,
            chunk_size=self.chunk_size,
            hdf5_dataset_name=self.hdf5_dataset_name,
        )

        # Split file paths into train, validation, and test sets
        generator = torch.Generator().manual_seed(self.rnd_seed)
        [train_dataset, validation_dataset, test_dataset] = random_split(
            whole_dataset,
            [self.train_proportion, self.validation_proportion, self.test_proportion],
            generator=generator,
        )
        print(f"Shape of item: {train_dataset.__getitem__(idx=5).shape}")
        return train_dataset, validation_dataset, test_dataset


class TimeSeriesDatasetSplitterSmall(DataSplitter):
    def __init__(
        self,
        train_proportion: int | float,
        validation_proportion: int | float = 0.0,
        test_proportion: int | float = 0.0,
        rnd_seed: Optional[int] = None,
        images_dataset: str = "data/Image_dataset_synthetic_64x64.pkl",
        name: Optional[str] = None,
    ) -> None:
        """Class for splitting of smaller datasets. Use this class in the pipeline if the
        entire dataset can fit into memory.

        Args:
            train_proportion (int | float): _description_
            validation_proportion (int | float, optional): _description_. Defaults to 0.0.
            test_proportion (int | float, optional): _description_. Defaults to 0.0.
            rnd_seed (Optional[int], optional): _description_. Defaults to None.
            images_dataset (str, optional): _description_.
                Defaults to "data/Image_dataset_synthetic_64x64.pkl".
            name (Optional[str], optional): _description_. Defaults to None.
        """
        super().__init__(train_proportion, validation_proportion, test_proportion, name)
        self.save_parameters(**self.locals2params(locals()))
        self.validation_proportion = 1 - train_proportion
        self.rnd_seed = rnd_seed
        self.images_dataset = images_dataset

    def get_or_load(self, dataset: Optional[pd.DataFrame] = None):
        """If the dataset is not given, load it from disk."""
        if dataset is None:
            print("WARNING: loading time series dataset from disk.")
            return pd.read_pickle(self.images_dataset)
        return dataset

    @monitor_exec
    def execute(self, dataset: Optional[pd.DataFrame] = None) -> Tuple:
        """Splits a dataset into train, validation and test splits.

        Args:
            dataset (pd.DataFrame): input dataset.

        Returns:
            Tuple: tuple of train, validation and test splits. Test is None.
        """
        dataset = self.get_or_load(dataset)

        # Convert data to torch
        df = dataset.map(lambda x: torch.tensor(x))

        # Divide Image dataset in main and aux channels. Note that df
        # generated in the section Generate Synthetic Dataset will always have
        # the main channel as its first column
        main_channel = list(df.columns)[0]
        aux_channels = list(df.columns)[1:]

        df_aux_all_2d = pd.DataFrame(df[aux_channels])
        df_main_all_2d = pd.DataFrame(df[main_channel])
        X_train_2d, X_test_2d, y_train_2d, y_test_2d = train_test_split(
            df_aux_all_2d,
            df_main_all_2d,
            test_size=self.validation_proportion,
            random_state=self.rnd_seed,
        )
        return (X_train_2d, y_train_2d), (X_test_2d, y_test_2d)


class TimeSeriesProcessorSmall(DataProcessor):
    def __init__(self, name: str | None = None) -> None:
        """Preprocesses small datasets that can fit into memory.

        Args:
            name (str | None, optional): Defaults to None.
        """
        super().__init__(name)
        self.save_parameters(**self.locals2params(locals()))

    @monitor_exec
    def execute(
        self, train_dataset: Tuple, validation_dataset: Tuple
    ) -> Tuple[torch.Tensor, torch.Tensor, None]:
        """Pre-process datasets: rearrange and normalize before training.

        Args:
            train_dataset (Tuple): training dataset.
            validation_dataset (Tuple): validation dataset.
            test_dataset (Any, optional): unused placeholder. Defaults to None.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, None]: train, validation, and
                test (placeholder) datasets. Ready to be used for training.
        """
        X_train_2d, y_train_2d = train_dataset
        X_test_2d, y_test_2d = validation_dataset

        # Name of the main channel (assuming it's in position 0)
        main_channel = list(y_train_2d.columns)[0]

        # TRAINING SET

        # # smaller dataset
        # signal_data_train_small_2d = torch.stack([
        #     torch.stack([y_train_2d[main_channel].iloc[i]])
        #     for i in range(100)
        # ])  # for i in range(y_train.shape[0])
        # aux_data_train_small_2d = torch.stack([
        #     torch.stack([X_train_2d.iloc[i][0], X_train_2d.iloc[i]
        #                 [1], X_train_2d.iloc[i][2]])
        #     for i in range(100)
        # ])  # for i in range(X_train.shape[0])

        # whole dataset
        signal_data_train_2d = torch.stack(
            [
                torch.stack([y_train_2d[main_channel].iloc[i]])
                for i in range(y_train_2d.shape[0])
            ]
        )
        aux_data_train_2d = torch.stack(
            [
                torch.stack(
                    [
                        X_train_2d.iloc[i][0],
                        X_train_2d.iloc[i][1],
                        X_train_2d.iloc[i][2],
                    ]
                )
                for i in range(X_train_2d.shape[0])
            ]
        )

        # concatenate torch.tensors
        train_data_2d = torch.cat([signal_data_train_2d, aux_data_train_2d], dim=1)
        # train_data_small_2d = torch.cat(
        #     [signal_data_train_small_2d, aux_data_train_small_2d], dim=1)

        # VALIDATION SET

        # # smaller dataset
        # signal_data_test_small_2d = torch.stack([
        #     torch.stack(
        #         [y_test_2d[main_channel].iloc[i]])
        #     for i in range(100)
        # ])  # for i in range(y_test.shape[0])
        # aux_data_test_small_2d = torch.stack([
        #     torch.stack(
        #         [X_test_2d.iloc[i][0], X_test_2d.iloc[i][1],
        #          X_test_2d.iloc[i][2]])
        #     for i in range(100)
        # ])  # for i in range(X_test.shape[0])

        # whole dataset
        signal_data_test_2d = torch.stack(
            [torch.stack([y_test_2d[main_channel].iloc[i]]) for i in range(y_test_2d.shape[0])]
        )
        aux_data_test_2d = torch.stack(
            [
                torch.stack([X_test_2d.iloc[i, 0], X_test_2d.iloc[i, 1], X_test_2d.iloc[i, 2]])
                for i in range(X_test_2d.shape[0])
            ]
        )

        test_data_2d = torch.cat([signal_data_test_2d, aux_data_test_2d], dim=1)
        # test_data_small_2d = torch.cat(
        #     [signal_data_test_small_2d, aux_data_test_small_2d], dim=1)

        # NORMALIZE
        train_data_2d = normalize_(train_data_2d)
        test_data_2d = normalize_(test_data_2d)

        print(f"Shape of training tensor: {train_data_2d.shape}")

        return TensorDataset(train_data_2d), TensorDataset(test_data_2d), None



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import dagger
from dagger import enum_type


@enum_type
class Stage(dagger.Enum):
    """Image distribution stage"""

    DEV = "DEV", "Pushed as itwinai-dev"
    PRODUCTION = "PRODUCTION", "Pushed as itwinai"
    CVMFS = "CVMFS", "Pushed as itwinai-cvmfs"


@enum_type
class MLFramework(dagger.Enum):
    TORCH = "TORCH"
    TENSORFLOW = "TENSORFLOW"



#!/opt/conda/bin/python
# -*- coding: utf-8 -*-
#
# D. Ciangottini
#
import asyncio
import os
import re
import sys
from subprocess import Popen

import asyncssh
from jupyterhub.singleuser import main

ssh_host = os.environ.get("JHUB_HOST")
ssh_url_port = os.environ.get("SSH_PORT")
username = os.environ.get("JUPYTERHUB_USER")
token = os.environ.get("JUPYTERHUB_API_TOKEN")

fwd_port = os.environ.get("FWD_PORT")


async def run_client():
    async with asyncssh.connect(
        host=ssh_host,
        port=int(ssh_url_port),
        username=username,
        password=token,
        known_hosts=None,
    ) as conn:
        conn.set_keepalive(interval=14.0, count_max=10)
        listener = await conn.forward_remote_port(
            "0.0.0.0",
            int(fwd_port),
            "0.0.0.0",
            int(fwd_port),
        )
        await listener.wait_closed()


if __name__ == '__main__':
    print("Connecting ssh...")
    loop = asyncio.get_event_loop()
    loop.create_task(run_client())

    print("Configuring Rucio extension...")
    p = Popen(['/usr/local/bin/setup.sh'])
    while p.poll() is None:
        pass

    print("Starting JLAB")
    sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])
    sys.exit(main())



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

"""Test distributed training strategies."""

from typing import Any
from unittest.mock import patch

import pytest
import torch
import torch.nn as nn
from torch.nn import Linear
from torch.optim import SGD, Optimizer
from torch.utils.data import DataLoader, Dataset, DistributedSampler

from itwinai.torch.distributed import (
    DeepSpeedStrategy,
    HorovodStrategy,
    TorchDDPStrategy,
    TorchDistributedStrategy,
)
from itwinai.torch.type import DistributedStrategyError, UninitializedStrategyError


class DummyDataset(Dataset):
    def __init__(self, size=100):
        self.size = size
        self.data = torch.randn(size, 10)
        self.targets = torch.randint(0, 2, (size,))

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        return self.data[idx], self.targets[idx]


class BaseTestDistributedStrategy:
    @pytest.fixture(scope="module")
    def strategy(self):
        """To be overridden by subclasses with the specific strategy."""
        raise NotImplementedError

    @pytest.fixture
    def simple_model(self):
        return Linear(10, 2)

    @pytest.fixture
    def optimizer(self, simple_model: nn.Module):
        return SGD(simple_model.parameters(), lr=0.01)

    @pytest.fixture
    def dataset(self):
        return DummyDataset()

    def test_cluster_properties(self, strategy: TorchDistributedStrategy):
        """Test that ranks and other simple properties are computed correctly."""
        assert strategy.is_initialized
        assert isinstance(strategy.device(), str)
        assert isinstance(strategy.is_main_worker, bool)
        assert strategy.global_world_size() >= 1
        assert strategy.local_world_size() >= 1
        assert strategy.global_rank() >= 0
        assert strategy.local_rank() >= 0

    def test_init_exceptions(
        self, strategy: TorchDistributedStrategy, simple_model: nn.Module, optimizer: Any
    ):
        """Check that the init method cannot be called twice and that the other methods raise
        and exception if called when the strategy is not initialized."""
        # Test re-initialization
        with pytest.raises(DistributedStrategyError) as init_exc:
            strategy.init()
            assert "already initialized" in init_exc.value

        # Test initialized flag
        strategy.is_initialized = False

        with pytest.raises(UninitializedStrategyError):
            strategy.distributed(simple_model, optimizer)
        with pytest.raises(UninitializedStrategyError):
            strategy.global_rank()
        with pytest.raises(UninitializedStrategyError):
            strategy.local_rank()
        with pytest.raises(UninitializedStrategyError):
            strategy.local_world_size()
        with pytest.raises(UninitializedStrategyError):
            strategy.global_world_size()
        with pytest.raises(UninitializedStrategyError):
            strategy.device()
        with pytest.raises(UninitializedStrategyError):
            strategy.is_main_worker
        x = torch.ones(2)
        with pytest.raises(UninitializedStrategyError):
            strategy.gather(x)
        with pytest.raises(UninitializedStrategyError):
            strategy.allgather_obj(x)
        with pytest.raises(UninitializedStrategyError):
            strategy.gather_obj(x)
        with pytest.raises(UninitializedStrategyError):
            strategy.clean_up()

        strategy.is_initialized = True

    def test_dataloader(self, strategy: TorchDistributedStrategy, dataset: Dataset):
        """Check that the dataloader is distributed correctly."""
        dataloader = strategy.create_dataloader(dataset, batch_size=16, shuffle=True)
        assert isinstance(dataloader, DataLoader)
        assert isinstance(dataloader.sampler, DistributedSampler)
        assert dataloader.sampler.rank == strategy.global_rank()
        assert dataloader.sampler.num_replicas == strategy.global_world_size()

    def test_gather_operations(self, strategy):
        """Test collective operations."""
        # Test tensor gather
        local_tensor = torch.tensor([strategy.global_rank()], device=strategy.device())
        gathered = strategy.gather(local_tensor)
        if strategy.is_main_worker:
            assert len(gathered) == strategy.global_world_size()
        else:
            assert gathered is None

        # Test tensor gather from CPU
        my_tensor = torch.ones(10) * strategy.global_rank()
        tensors = strategy.gather(my_tensor, dst_rank=0)
        if strategy.is_main_worker:
            assert torch.stack(tensors).sum() == sum(range(strategy.global_world_size())) * 10
        else:
            assert tensors is None

        # Test object gather
        local_obj = {"rank": strategy.global_rank()}
        gathered_obj = strategy.gather_obj(local_obj)
        if strategy.is_main_worker:
            assert len(gathered_obj) == strategy.global_world_size()
        else:
            assert gathered_obj is None

        # Test allgather
        all_gathered = strategy.allgather_obj(local_obj)
        assert len(all_gathered) == strategy.global_world_size()

    def test_ranks(self, strategy):
        """Check that the values of the ranks are as expected."""
        nnodes = strategy.global_world_size() // strategy.local_world_size()

        # Test local ranks
        lranks = strategy.gather_obj(strategy.local_rank(), dst_rank=0)
        if strategy.is_main_worker:
            assert len(lranks) == strategy.global_world_size()
            assert sum(lranks) == sum(range(strategy.local_world_size())) * nnodes
        else:
            assert lranks is None

        # Test global ranks
        granks = strategy.gather_obj(strategy.global_rank(), dst_rank=0)
        if strategy.is_main_worker:
            assert len(granks) == strategy.global_world_size()
            assert sum(granks) == sum(range(strategy.global_world_size()))
        else:
            assert granks is None


@pytest.mark.hpc
@pytest.mark.torch_dist
class TestTorchDDPStrategy(BaseTestDistributedStrategy):
    @pytest.fixture(scope="module")
    def strategy(self, ddp_strategy) -> TorchDDPStrategy:
        return ddp_strategy

    def test_init(self, strategy: TorchDDPStrategy):
        """Test specific initialization of TorchDDPStrategy."""
        assert strategy.backend in ["nccl", "gloo"]

        # Test initialization
        init_path = "torch.distributed.init_process_group"
        with patch(init_path, autospec=True) as mock_init_torch:
            strategy = TorchDDPStrategy(
                backend="nccl" if torch.cuda.is_available() else "gloo"
            )
            strategy.init()
            mock_init_torch.assert_called_once()

    def test_distributed_model(
        self, strategy: TorchDDPStrategy, simple_model: nn.Module, optimizer: Optimizer
    ):
        """Test NN model distribution."""
        from torch.nn.parallel import DistributedDataParallel

        dist_model, dist_optimizer, _ = strategy.distributed(simple_model, optimizer)
        assert isinstance(dist_model, DistributedDataParallel)
        assert hasattr(dist_model, "module") or isinstance(dist_model, nn.Module)
        assert isinstance(dist_optimizer, Optimizer)


@pytest.mark.hpc
@pytest.mark.deepspeed_dist
class TestDeepSpeedStrategy(BaseTestDistributedStrategy):
    @pytest.fixture(scope="module")
    def strategy(self, deepspeed_strategy) -> DeepSpeedStrategy:
        return deepspeed_strategy

    def test_init(self, strategy: DeepSpeedStrategy):
        """Test specific initialization of DeepSpeedStrategy."""
        assert strategy.backend in ["nccl", "gloo", "mpi"]
        assert hasattr(
            strategy, "deepspeed"
        ), "Lazy import of deepspeed not found in strategy class."

        # Test initialization
        init_path = "deepspeed.init_distributed"
        with patch(init_path, autospec=True) as mock_init_ds:
            strategy = DeepSpeedStrategy(
                backend="nccl" if torch.cuda.is_available() else "gloo"
            )
            strategy.init()
            mock_init_ds.assert_called_once()

    def test_distributed_model(
        self, strategy: DeepSpeedStrategy, simple_model: nn.Module, optimizer: Optimizer
    ):
        """Test NN model distribution."""
        from deepspeed.runtime.engine import DeepSpeedEngine

        ds_config = {
            "train_batch_size": 16,
            "fp16": {"enabled": False},
            "optimizer": {"type": "SGD", "params": {"lr": 0.001}},
        }

        dist_model, dist_optimizer, _ = strategy.distributed(
            simple_model, optimizer=optimizer, config=ds_config
        )
        assert hasattr(dist_model, "module")
        assert isinstance(dist_model, DeepSpeedEngine)
        assert dist_optimizer is not None
        assert isinstance(dist_optimizer, Optimizer)


@pytest.mark.hpc
@pytest.mark.horovod_dist
class TestHorovodStrategy(BaseTestDistributedStrategy):
    @pytest.fixture(scope="module")
    def strategy(self, horovod_strategy) -> HorovodStrategy:
        return horovod_strategy

    def test_init(self, strategy):
        assert strategy.is_initialized
        assert hasattr(strategy, "hvd"), "Lazy import of horovod not found in strategy class."

        # Test initialization
        init_path = "horovod.torch.init"
        with patch(init_path, autospec=True) as mock_init_ds:
            strategy = HorovodStrategy()
            strategy.init()
            mock_init_ds.assert_called_once()

    def test_distributed_model(self, strategy, simple_model, optimizer):
        dist_model, dist_optimizer, _ = strategy.distributed(
            simple_model, optimizer=optimizer, op=strategy.hvd.Average
        )
        assert isinstance(dist_model, nn.Module)
        assert isinstance(dist_optimizer, Optimizer)
        assert hasattr(
            dist_optimizer, "synchronize"
        ), "synchronize() method not found for Horovod optimizer"



# --------------------------------------------------------------------------------------
# Part of the interTwin Project: https://www.intertwin.eu/
#
# Created by: Matteo Bunino
#
# Credit:
# - Matteo Bunino <matteo.bunino@cern.ch> - CERN
# --------------------------------------------------------------------------------------

import copy
import logging
import re
import time
from typing import Dict, List, Optional

import yaml
from kubernetes import client, config
from kubernetes.client.rest import ApiException

# Template with default values
POD_TEMPLATE = {
    "apiVersion": "v1",
    "kind": "Pod",
    "metadata": {
        "name": "pod-template",
        "namespace": "default",
        "annotations": {
            "slurm-job.vk.io/flags": (
                "-p gpu --gres=gpu:1 --ntasks-per-node=1 --nodes=1 --time=00:10:00"
            ),
            "slurm-job.vk.io/pre-exec": "ls -la",
        },
    },
    "spec": {
        "containers": [
            {
                "name": "busy-container",
                "image": "/ceph/hpc/data/st2301-itwin-users/cern/hello-world-image.sif",
                "command": ["/bin/sh", "-c"],
                "args": ["sleep $(( 60 * 1 ))"],
                "imagePullPolicy": "Always",
                "resources": {
                    "limits": {"cpu": "48", "memory": "150Gi"},
                    "requests": {"cpu": "4", "memory": "20Gi"},
                },
            }
        ],
        "restartPolicy": "Always",
        "nodeSelector": {"kubernetes.io/hostname": "vega-new-vk"},
        "tolerations": [
            {"key": "virtual-node.interlink/no-schedule", "operator": "Exists"},
            {
                "effect": "NoExecute",
                "key": "node.kubernetes.io/not-ready",
                "operator": "Exists",
                "tolerationSeconds": 300,
            },
            {
                "effect": "NoExecute",
                "key": "node.kubernetes.io/unreachable",
                "operator": "Exists",
                "tolerationSeconds": 300,
            },
        ],
    },
}


def validate_pod_name(name: str) -> None:
    """Validates if a given string matches the Kubernetes pod naming convention.

    Args:
        name (str): The pod name to validate.

    Raises:
        ValueError: If the name does not match the required pattern.
    """
    pattern = r"[a-z0-9]([-a-z0-9]*[a-z0-9])?(\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*"
    if not re.fullmatch(pattern, name):
        raise ValueError(
            f"The name '{name}' is invalid for a Kubernetes pod. "
            "Pod names must match the pattern: "
            "'[a-z0-9]([-a-z0-9]*[a-z0-9])?(\\.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*'"
        )


def create_pod_manifest(
    name: str = None,
    annotations: Dict = None,
    container_name: str = None,
    image_path: str = None,
    cmd_args: List[str] = None,
    resources: Dict = None,
):
    """Creates a pod manifest with optional parameters to override default values
    for pod configuration.

    Args:
        name (str, optional): Name of the pod (metadata.name).
        annotations (Dict, optional): Annotations for the pod (metadata.annotations).
        container_name (str, optional): Name of the container (spec.containers[0].name).
        image_path (str, optional): Path to the container image
            (spec.containers[0].image).
        cmd_args (List[str], optional): Command arguments for the container
            (spec.containers[0].args).
        resources (Dict, optional): Resource requests and limits for the container
            (spec.containers[0].resources).

    Returns:
        Dict: A dictionary representing the pod manifest.
    """

    # Deep copy the template to avoid modifying the original
    pod_manifest = copy.deepcopy(POD_TEMPLATE)

    # Set metadata fields
    if name:
        pod_manifest["metadata"]["name"] = name
    if annotations:
        pod_manifest["metadata"]["annotations"].update(annotations)

    # Set container-specific fields
    container = pod_manifest["spec"]["containers"][0]
    if container_name:
        container["name"] = container_name
    if image_path:
        container["image"] = image_path
    if cmd_args:
        container["args"] = cmd_args
    if resources:
        container["resources"] = resources

    return pod_manifest


def load_kube_config_from_string(kubeconfig_str: str):
    kubeconfig_dict = yaml.safe_load(kubeconfig_str)
    config.load_kube_config_from_dict(kubeconfig_dict)


def load_pod_manifest_from_yaml(file_path: str) -> Dict:
    """Load a pod manifest from a YAML file."""
    with open(file_path, "r") as f:
        pod_manifest = yaml.safe_load(f)
    return pod_manifest


def create_pod(api_instance: client.CoreV1Api, namespace: str, pod_manifest: Dict):
    try:
        api_response = api_instance.create_namespaced_pod(
            namespace=namespace, body=pod_manifest
        )
        print(f"Pod created. Status: {api_response.status.phase}")
    except ApiException as e:
        print(f"Exception when creating pod: {e}")


def check_pod_status(api_instance: client.CoreV1Api, namespace: str, pod_name: str):
    try:
        pod = api_instance.read_namespaced_pod(name=pod_name, namespace=namespace)
        return pod.status.phase
    except ApiException as e:
        print(f"Exception when checking pod status: {e}")
        return None


def get_pod_logs_insecure(api_instance: client.CoreV1Api, namespace: str, pod_name: str):
    """Fetch logs for the specified pod with insecure TLS settings."""
    try:
        log_response = api_instance.read_namespaced_pod_log(
            name=pod_name, namespace=namespace, insecure_skip_tls_verify_backend=True
        )
        return log_response
    except ApiException as e:
        print(f"Exception when retrieving pod logs: {e}")
        return None


def delete_pod(api_instance: client.CoreV1Api, namespace: str, pod_name: str):
    """Delete a pod by its name in a specified namespace."""
    try:
        api_response = api_instance.delete_namespaced_pod(name=pod_name, namespace=namespace)
        print(f"Pod '{pod_name}' deleted. Status: {api_response.status}")
    except ApiException as e:
        print(f"Exception when deleting pod: {e}")


def submit_job(
    kubeconfig_str: str, pod_manifest: Dict, verbose: bool = False
) -> Optional[str]:
    """Submit a pod to some kubernetes cluster.

    Args:
        kubeconfig_str (str): kubeconfig for kubernetes in string format.
        pod_manifest (Dict): manifest of the pod.
        verbose (bool, optional): whether to print pod logs every second. Defaults to False.

    Returns:
        Optional[str]: final pod status.
    """
    load_kube_config_from_string(kubeconfig_str)

    # Initialize the CoreV1Api for managing pods
    v1 = client.CoreV1Api()

    namespace = "default"
    pod_name = pod_manifest["metadata"]["name"]

    # Kill existing pod, if present
    status = check_pod_status(v1, namespace, pod_name)
    if status:
        logging.warning(f"Pod {pod_name} already existed... Deleting it before continuing.")
        delete_pod(v1, namespace, pod_name)
        while status is not None:
            time.sleep(1)
            status = check_pod_status(v1, namespace, pod_name)

    # Create the pod in the specified namespace
    create_pod(v1, namespace, pod_manifest)
    status = None
    logs = None
    try:
        # Check the pod status
        status = check_pod_status(v1, namespace, pod_name)
        print(f"Pod status: {status}")
        print("=" * 100)
        print("Waiting for pod completion...")

        # Get pod logs with insecure TLS settings if the pod has completed its
        # initialization
        while status in ["Running", "Pending"]:
            time.sleep(1)
            status = check_pod_status(v1, namespace, pod_name)
            logs = get_pod_logs_insecure(v1, namespace, pod_name)
            if verbose:
                print(f"Pod status: {status}")
                print(f"Pod logs:\n{logs}")
    finally:
        print("=" * 100)
        print(f"Pod status: {status}")
        print(f"Pod logs:\n{logs}")
        delete_pod(v1, namespace, pod_name)
    return status
